import gc
import os
import pickle

import numpy as np
import pandas as pd
import sklearn
import tensorflow as tf
from IPython.core.debugger import set_trace
from sklearn import metrics
from tensorflow.keras.preprocessing import sequence, text

import kaggle_runner.utils.kernel_utils as utils
from kaggle_runner import defaults

MAX_LEN = 180  # used in lstm definition ....


# from tqdm import tqdm
# tqdm.pandas()


MODEL_NAME = "lstm"
VAL_ERR_COLUMN = "err"
EMBEDDING_FILES = [
    "../input/glove840b300dtxt/glove.840B.300d.txt",
    "../input/fasttext-crawl-300d-2m/crawl-300d-2M.vec",
]
# only read first 220 words in the comment sentence
IDENTITY_COLUMNS = [
    "male",
    "female",
    "homosexual_gay_or_lesbian",
    "christian",
    "jewish",
    "muslim",
    "black",
    "white",
    "psychiatric_or_mental_illness",
]  # features, meaning
TARGET_COLUMN = "target"
AUX_COLUMNS = [
    TARGET_COLUMN,
    "severe_toxicity",
    "obscene",
    "identity_attack",
    "insult",
    "threat",
    "sexual_explicit",
]
TEXT_COLUMN = "comment_text"
# CHARS_TO_REMOVE = '!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n“”’\'∞θ÷α•à−β∅³π‘₹´°£€\×™√²—'
DATA_FILE_FLAG = ".tf_record_saved"


def maybe_download(url):
    # By default the file at the url origin is downloaded to the cache_dir ~/.keras,
    # placed in the cache_subdir datasets, and given the filename fname
    path = tf.keras.utils.get_file(url.split("/")[-1], url)

    return path


def train_input_fn_general(
        features,
        labels,
        batch_size,
        cv,
        split_id=None,
        n_splits=None,
        handle_large=False,
        shuffle_size=None,
):
    # for boost tree, need to prepare feature columns
    # 2048? columns, all float

    if cv:
        return _input_fn_general(
            features,
            labels,
            batch_size,
            split_id=split_id,
            n_splits=n_splits,
            cv_train=True,
            handle_large=handle_large,
        )
    else:
        return _input_fn_general(features, labels, batch_size, cv=False)


def eval_input_fn_general(
    features, labels, batch_size, cv, split_id=None, n_splits=None
):
    if cv:
        return _input_fn_general(
            features,
            labels,
            batch_size,
            with_y=True,
            repeat=False,
            shuffle=False,
            split_id=split_id,
            n_splits=n_splits,
            cv_train=False,
        )
    else:
        return _input_fn_general(
            features,
            labels,
            batch_size,
            with_y=True,
            repeat=False,
            shuffle=False,
            cv=False,
        )


def pred_input_fn_general(features, batch_size):
    return _input_fn_general(
        features, None, batch_size, with_y=False, repeat=False, shuffle=False, cv=False
    )


def _input_fn_general(
        features,
        labels,
        batch_size,
        with_y=True,
        repeat=True,
        shuffle=True,
        split_id=-1,
        n_splits=10,
        cv=True,
        cv_train=True,
        to_dict=False,
        handle_large=False,
        shuffle_size=None,
):
    # for these, we will need to extract all the points before:
    """_input_fn_general is used for (boot tree)?

    :param features:
    :param labels:
    :param batch_size:
    :param with_y:
    :param repeat:
    :param shuffle:
    :param split_id:
    :param n_splits:
    :param cv:
    :param cv_train: cv train or evaluation
    :return:
    """

    def _to_dict(f):
        # first to pandas data frame
        df = pd.DataFrame(f, columns=[str(i)
                                      for i in range(features.shape[-1])])

        return dict(df)

    if to_dict:
        features = _to_dict(features)

    if handle_large:
        # features_placeholder = tf.placeholder(features.dtype, features.shape)
        # labels_placeholder = tf.placeholder(labels.dtype, labels.shape)
        features = tf.keras.backend.placeholder(
            dtype=features.dtype, shape=features.shape
        )
        labels = tf.keras.backend.placeholder(
            dtype=labels.dtype, shape=labels.shape)

    if with_y:
        ds = tf.data.Dataset.from_tensor_slices((features, labels))
    else:
        ds = tf.data.Dataset.from_tensor_slices(features)

    if cv:
        assert split_id >= 0 and n_splits > 1 and split_id < n_splits

        if cv_train:
            ds = [ds.shard(n_splits, i) for i in range(n_splits)]

            shards_cross = [
                ds[val_id] for val_id in range(n_splits) if val_id != split_id
            ]

            ds = shards_cross[0]

            for t in shards_cross[1:]:
                ds = ds.concatenate(t)

            if shuffle:
                if handle_large:
                    ds = ds.shuffle(
                        buffer_size=shuffle_size
                    )  # just memory is not enough ...
                else:
                    ds = ds.shuffle(
                        buffer_size=int(
                            len(labels) * (n_splits - 1) / n_splits)
                    )  # just memory is not enough ...
        else:  # cv evaluate, no need to shuffle
            ds = ds.shard(n_splits, split_id)
    else:
        if shuffle:
            if handle_large:
                ds = ds.shuffle(buffer_size=shuffle_size)
            else:
                ds = ds.shuffle(buffer_size=len(labels))
    # after shuffle, we do cross validtation split

    # taken from Dan, https://stackoverflow.com/questions/39748660/how-to-perform-k-fold-cross-validation-with-tensorflow
    # will need to append id, then remove the id?
    # -> no need, we just split to 5 shards, then rearrange these shards

    if repeat:
        ds = ds.repeat()
    # Batch the examples
    assert batch_size is not None, "batch_size must not be None"
    # Return the dataset.

    return ds.batch(batch_size).prefetch(1)


def _load_embeddings_only_for_fasttext_crawl_avg(path):
    """

    :param path: large embedding file
    :return: np array of
    """
    with open(path, "r") as f:
        for i, line in enumerate(f):
            if i == 0:
                continue  # for crawl the first line is "2000000 300"
            # print(line)
            hidden_dim = (
                len(line.split(" ")) -
                2 if line.strip().split(" ")[-1] == "\n" else 1
            )  # for crawl, the '\n' is there to
            # print(hidden_dim)

            break

    vec_sum_place = np.zeros(hidden_dim, dtype=np.float32)
    # print(vec_sum_place.shape)
    with open(path, "r") as f:
        for i, line in enumerate(f):
            if i == 0:
                continue
            # try:
            nums = line.split(" ")[1:]

            if nums[-1] == "\n":
                nums = nums[:-1]
            vec_sum_place = vec_sum_place + np.array(
                [float(num) for num in nums], dtype=np.float32
            )
            # except ValueError as e:
            #    print(line)
            #    print(len(line.strip().split(' ')))
            #    print(e)
            #    raise e
    avg_vec = vec_sum_place / i  # i = 0 is the header line

    return avg_vec


# Define bias metrics, then evaluate our new model for bias using the validation set predictions
SUBGROUP_AUC = "subgroup_auc"
BPSN_AUC = "bpsn_auc"  # stands for background positive, subgroup negative
BNSP_AUC = "bnsp_auc"  # stands for background negative, subgroup positive
OVERALL_AUC = "overall_auc"


class TargetDistAnalyzer:
    def __init__(self, target):
        self.df = (
            target  # df with target, with identity information, then we can sort it out
        )
        self.discretizer = sklearn.preprocessing.KBinsDiscretizer(
            10 + 1, encode="ordinal", strategy="kmeans"
        )
        self.discretizer.fit(target[TARGET_COLUMN].values.reshape(-1, 1))
        self.id_used_in_train = None

    def get_distribution(self, target_data):
        """
        target_data: pandas series, need to get index, so need series, in the series, values are target (prediction)

        :return: (type, cnt number, frequency, index) pair list for this distribution
        """
        dst = []

        y_t = self.discretizer.transform(target_data.values.reshape(-1, 1))
        uniq_elements, element_counts = np.unique(y_t, return_counts=True)
        all_counts = len(y_t)

        for i, e in enumerate(uniq_elements):
            dst.append(
                (
                    e,
                    element_counts[i],
                    element_counts[i] / all_counts,
                    target_data.loc[(y_t == e).ravel()].index,
                )
            )

        return dst

    def get_err_distribution(self, err_data, val_mask):
        """

        :param err_data: series, contain both target values and error values
        :return: (type, cnt number, frequency, index) pair list for this distribution
        """
        dstr = {}

        df = err_data

        for g in IDENTITY_COLUMNS:
            val_subgroup_idx = df.loc[val_mask & (df[g] > 0.5)].index

            val_subgroup = df.loc[val_subgroup_idx,
                                  [TARGET_COLUMN, VAL_ERR_COLUMN]]

            dst = self.get_distribution(
                val_subgroup[TARGET_COLUMN]
            )  # could use continuous data, might be helpful so calculate belongs (fuzzy logic)

            err_mean_in_split = [
                val_subgroup.loc[d[3], VAL_ERR_COLUMN].mean() for d in dst
            ]
            dstr[g] = (dst, err_mean_in_split)

        return dstr

    def get_distribution_overall(self):
        return self.get_distribution(self.df[TARGET_COLUMN])

    def get_distribution_subgroups(self):
        dstr = {}

        for g in IDENTITY_COLUMNS:
            dstr[g] = self.get_distribution(
                self.df[self.df[g +
                                "_in_train"].fillna(0.0) >= 0.5][TARGET_COLUMN]
            )  # could use continuous data, might be helpful so calculate belongs (fuzzy logic)

        return dstr


class BiasBenchmark:
    def __init__(self, train_df_id_na_dropped, threshold=0.5):
        # we can use this to divide subgroups(beside, we need to calculate only the 45000, with identities
        self.threshold = threshold
        self.validate_df = BiasBenchmark.copy_convert_dataframe_to_bool(
            train_df_id_na_dropped[IDENTITY_COLUMNS +
                                   [TARGET_COLUMN, TEXT_COLUMN]],
            threshold,
        )

    @staticmethod
    def compute_auc(y_true, y_pred):
        try:
            return metrics.roc_auc_score(y_true, y_pred)
        except ValueError:
            return np.nan

    @staticmethod
    def compute_predict_estimator(y_true, y_pred, y_pred_continuous=None):
        assert pd.core.dtypes.common.is_dtype_equal(
            y_true.dtype, np.dtype("bool"))

        pos_cnt = sum(y_true)
        neg_cnt = len(y_true) - pos_cnt

        pred_pos_cnt = sum(y_pred > 0.5)
        pred_neg_cnt = len(y_pred) - pred_pos_cnt

        at_predict = y_pred[
            y_true
        ]  # predictions for True actually, might predict to 0.4(bad)
        af_predict = y_pred[~y_true]

        if y_pred_continuous is None:
            return (
                (int(neg_cnt), int(pos_cnt), pos_cnt / (neg_cnt + pos_cnt)),
                (int(pred_neg_cnt), int(pred_pos_cnt)),
                (len(af_predict), np.mean(af_predict), np.std(af_predict)),
                (len(at_predict), np.mean(at_predict), np.std(at_predict)),
            )
        else:
            at_cont_val = y_pred_continuous[y_true]
            af_cont_val = y_pred_continuous[~y_true]

            return (
                (int(neg_cnt), int(pos_cnt), pos_cnt / (neg_cnt + pos_cnt)),
                (int(pred_neg_cnt), int(pred_pos_cnt)),
                (len(af_predict), np.mean(af_predict), np.std(af_predict)),
                (len(at_predict), np.mean(at_predict), np.std(at_predict)),
                (len(af_cont_val), np.mean(af_cont_val), np.std(af_cont_val)),
                (len(at_cont_val), np.mean(at_cont_val), np.std(at_cont_val)),
            )

    @staticmethod
    def compute_subgroup_classify_detail(df, subgroup, label, model_name):
        """
        Compute AUC for spefic subgroup
        :param df: dataframe which contains predictions for all subgroups
        :param subgroup: compute AUC for this subgroup
        :param label: target column name
        :param model_name:
        :return: just mean/std, for pos, neg, then we just use this information to make shift
        """
        subgroup_examples = df[
            df[subgroup]
        ]  # innter df[subgroup] will get out boolean list, which is used to select this
        # subgroup examples

        if label == TARGET_COLUMN:
            return BiasBenchmark.compute_predict_estimator(
                subgroup_examples[label],
                subgroup_examples[model_name],
                y_pred_continuous=subgroup_examples[label + "_orig"],
            )
        else:
            return BiasBenchmark.compute_predict_estimator(
                subgroup_examples[label], subgroup_examples[model_name]
            )

    @staticmethod
    def compute_subgroup_auc(df, subgroup, label, model_name):
        """
        Compute AUC for spefic subgroup. AUC only cares about ordering, not threshold
        :param df: dataframe which contains predictions for all subgroups
        :param subgroup: compute AUC for this subgroup
        :param label: target column name
        :param model_name:
        :return: auc score
        """
        subgroup_examples = df[
            df[subgroup]
        ]  # innter df[subgroup] will get out boolean list, which is used to select this
        # subgroup examples

        return BiasBenchmark.compute_auc(
            subgroup_examples[label], subgroup_examples[model_name]
        )

    @staticmethod
    def compute_bpsn_classify_detail(df, subgroup, label, model_name):
        """Computes the AUC of the within-subgroup negative examples and the background positive examples."""
        subgroup_negative_examples = df[df[subgroup] & ~df[label]]
        non_subgroup_positive_examples = df[
            ~df[subgroup] & df[label]
        ]  # background positive
        examples = subgroup_negative_examples.append(
            non_subgroup_positive_examples)

        # this example is background True positive, with subgroup True negative, and see our model's prediction's performance

        return BiasBenchmark.compute_predict_estimator(
            examples[label], examples[model_name]
        )

    @staticmethod
    def compute_bpsn_auc(df, subgroup, label, model_name):
        """Computes the AUC of the within-subgroup negative examples and the background positive examples."""
        subgroup_negative_examples = df[df[subgroup] & ~df[label]]
        non_subgroup_positive_examples = df[~df[subgroup] & df[label]]
        examples = subgroup_negative_examples.append(
            non_subgroup_positive_examples)

        return BiasBenchmark.compute_auc(examples[label], examples[model_name])

    @staticmethod
    def compute_bnsp_classify_detail(df, subgroup, label, model_name):
        """Computes the AUC of the within-subgroup positive examples and the background negative examples."""
        subgroup_positive_examples = df[df[subgroup] & df[label]]
        non_subgroup_negative_examples = df[~df[subgroup] & ~df[label]]
        examples = subgroup_positive_examples.append(
            non_subgroup_negative_examples)

        return BiasBenchmark.compute_predict_estimator(
            examples[label], examples[model_name]
        )

    @staticmethod
    def compute_bnsp_auc(df, subgroup, label, model_name):
        """Computes the AUC of the within-subgroup positive examples and the background negative examples."""
        subgroup_positive_examples = df[df[subgroup] & df[label]]
        non_subgroup_negative_examples = df[~df[subgroup] & ~df[label]]
        examples = subgroup_positive_examples.append(
            non_subgroup_negative_examples)

        return BiasBenchmark.compute_auc(examples[label], examples[model_name])

    @staticmethod
    def compute_bias_metrics_for_model(
        dataset, subgroups, model, label_col, include_asegs=False
    ):
        """
        Computes per-subgroup metrics for all subgroups and one model.
        # bias_metrics_df = BiasBenchmark.compute_bias_metrics_for_model(validate_df, IDENTITY_COLUMNS, MODEL_NAME, TARGET_COLUMN)
        :param dataset: prediction result
        :param subgroups: all group names
        :param model: my model name
        :param label_col: target column name
        :param include_asegs: ?
        :return:
        """
        records = []
        subgroup_distribution = []

        for subgroup in subgroups:
            record = {
                "subgroup": subgroup,
                "subgroup_size": len(dataset[dataset[subgroup]]),
            }
            record_pn = {
                "subgroup": subgroup,
                "subgroup_size": len(dataset[dataset[subgroup]]),
            }
            record[SUBGROUP_AUC] = BiasBenchmark.compute_subgroup_auc(
                dataset, subgroup, label_col, model
            )
            record[BPSN_AUC] = BiasBenchmark.compute_bpsn_auc(
                dataset, subgroup, label_col, model
            )
            record[BNSP_AUC] = BiasBenchmark.compute_bnsp_auc(
                dataset, subgroup, label_col, model
            )

            record_pn[SUBGROUP_AUC] = BiasBenchmark.compute_subgroup_classify_detail(
                dataset, subgroup, label_col, model
            )
            record_pn[BPSN_AUC] = BiasBenchmark.compute_bpsn_classify_detail(
                dataset, subgroup, label_col, model
            )
            record_pn[BNSP_AUC] = BiasBenchmark.compute_bnsp_classify_detail(
                dataset, subgroup, label_col, model
            )

            records.append(record)
            subgroup_distribution.append(record_pn)

        return (
            pd.DataFrame(records).sort_values("subgroup_auc", ascending=True),
            subgroup_distribution,
        )

    # Calculate the final score
    @staticmethod
    def calculate_overall_auc(df, model_name):
        true_labels = df[TARGET_COLUMN]
        predicted_labels = df[model_name]
        try:
            return metrics.roc_auc_score(true_labels, predicted_labels)
        except ValueError:
            utils.logger("might be wrong")

            return 0.5  # bad

    @staticmethod
    def calculate_overall_auc_distribution(df, model_name):
        info = dict()
        info[OVERALL_AUC] = BiasBenchmark.calculate_overall_auc(df, model_name)
        info["subgroup"] = "overall"
        info["subgroup_size"] = len(df)
        info["distribution"] = BiasBenchmark.compute_predict_estimator(
            df[TARGET_COLUMN],
            df[model_name],
            y_pred_continuous=df[TARGET_COLUMN + "_orig"],
        )

        return info

    @staticmethod
    def power_mean(series, p):
        total = sum(np.power(series, p))

        return np.power(total / len(series), 1 / p)

    @staticmethod
    def get_final_metric(bias_df, overall_auc, POWER=-5, OVERALL_MODEL_WEIGHT=0.25):
        bias_metrics_on_subgroups = [
            BiasBenchmark.power_mean(bias_df[SUBGROUP_AUC], POWER),
            BiasBenchmark.power_mean(bias_df[BPSN_AUC], POWER),
            BiasBenchmark.power_mean(bias_df[BNSP_AUC], POWER),
        ]
        bias_score = np.average(bias_metrics_on_subgroups)

        return (
            (OVERALL_MODEL_WEIGHT * overall_auc)
            + ((1 - OVERALL_MODEL_WEIGHT) * bias_score),
            bias_metrics_on_subgroups,
        )

    @staticmethod
    def convert_to_bool(df, col_name, threshold=0.5, keep_original=False):
        if keep_original:
            df[col_name + "_orig"] = df[col_name]
        df[col_name] = np.where(df[col_name] >= threshold, True, False)

    @staticmethod
    def copy_convert_dataframe_to_bool(df, threshold):
        bool_df = df.copy()

        for col in IDENTITY_COLUMNS:
            BiasBenchmark.convert_to_bool(bool_df, col)
        BiasBenchmark.convert_to_bool(
            bool_df, TARGET_COLUMN, threshold, keep_original=True
        )

        return bool_df

    def calculate_benchmark(self, pred=None, validate_df=None, model_name=MODEL_NAME):
        """

        :param pred:
        :param model_name:
        :return: final metric score, bias auc for subgroups, subgroup classification distribution details, overall auc
        """

        if validate_df is None:
            print(
                "In caculating benchmark, the validate_df passed in None, use default validate_df, with given pred"
            )
            validate_df = self.validate_df
            assert validate_df.shape[0] == len(pred)
            validate_df[model_name] = pred  # prediction

        print("In caculating benchmark...")
        validate_df = BiasBenchmark.copy_convert_dataframe_to_bool(
            validate_df[IDENTITY_COLUMNS +
                        [TARGET_COLUMN, TEXT_COLUMN, model_name]],
            self.threshold,
        )
        (
            bias_metrics_df,
            subgroup_distribution,
        ) = BiasBenchmark.compute_bias_metrics_for_model(
            validate_df, IDENTITY_COLUMNS, model_name, TARGET_COLUMN
        )
        overall_auc_dist = BiasBenchmark.calculate_overall_auc_distribution(
            validate_df, model_name
        )
        final_score, score_comp = BiasBenchmark.get_final_metric(
            bias_metrics_df, overall_auc_dist[OVERALL_AUC]
        )

        return (
            final_score,
            score_comp,
            bias_metrics_df,
            subgroup_distribution,
            overall_auc_dist,
        )


# embedding vocab
class EmbeddingHandler:
    contraction_mapping = {
        "Trump's": "trump is",
        "'cause": "because",
        ",cause": "because",
        ";cause": "because",
        "ain't": "am not",
        "ain,t": "am not",
        "ain;t": "am not",
        "ain´t": "am not",
        "ain’t": "am not",
        "aren't": "are not",
        "aren,t": "are not",
        "aren;t": "are not",
        "aren´t": "are not",
        "aren’t": "are not",
        "can't": "cannot",
        "can't've": "cannot have",
        "can,t": "cannot",
        "can,t,ve": "cannot have",
        "can;t": "cannot",
        "can;t;ve": "cannot have",
        "can´t": "cannot",
        "can´t´ve": "cannot have",
        "can’t": "cannot",
        "can’t’ve": "cannot have",
        "could've": "could have",
        "could,ve": "could have",
        "could;ve": "could have",
        "couldn't": "could not",
        "couldn't've": "could not have",
        "couldn,t": "could not",
        "couldn,t,ve": "could not have",
        "couldn;t": "could not",
        "couldn;t;ve": "could not have",
        "couldn´t": "could not",
        "couldn´t´ve": "could not have",
        "couldn’t": "could not",
        "couldn’t’ve": "could not have",
        "could´ve": "could have",
        "could’ve": "could have",
        "didn't": "did not",
        "didn,t": "did not",
        "didn;t": "did not",
        "didn´t": "did not",
        "didn’t": "did not",
        "doesn't": "does not",
        "doesn,t": "does not",
        "doesn;t": "does not",
        "doesn´t": "does not",
        "doesn’t": "does not",
        "don't": "do not",
        "don,t": "do not",
        "don;t": "do not",
        "don´t": "do not",
        "don’t": "do not",
        "hadn't": "had not",
        "hadn't've": "had not have",
        "hadn,t": "had not",
        "hadn,t,ve": "had not have",
        "hadn;t": "had not",
        "hadn;t;ve": "had not have",
        "hadn´t": "had not",
        "hadn´t´ve": "had not have",
        "hadn’t": "had not",
        "hadn’t’ve": "had not have",
        "hasn't": "has not",
        "hasn,t": "has not",
        "hasn;t": "has not",
        "hasn´t": "has not",
        "hasn’t": "has not",
        "haven't": "have not",
        "haven,t": "have not",
        "haven;t": "have not",
        "haven´t": "have not",
        "haven’t": "have not",
        "he'd": "he would",
        "he'd've": "he would have",
        "he'll": "he will",
        "he's": "he is",
        "he,d": "he would",
        "he,d,ve": "he would have",
        "he,ll": "he will",
        "he,s": "he is",
        "he;d": "he would",
        "he;d;ve": "he would have",
        "he;ll": "he will",
        "he;s": "he is",
        "he´d": "he would",
        "he´d´ve": "he would have",
        "he´ll": "he will",
        "he´s": "he is",
        "he’d": "he would",
        "he’d’ve": "he would have",
        "he’ll": "he will",
        "he’s": "he is",
        "how'd": "how did",
        "how'll": "how will",
        "how's": "how is",
        "how,d": "how did",
        "how,ll": "how will",
        "how,s": "how is",
        "how;d": "how did",
        "how;ll": "how will",
        "how;s": "how is",
        "how´d": "how did",
        "how´ll": "how will",
        "how´s": "how is",
        "how’d": "how did",
        "how’ll": "how will",
        "how’s": "how is",
        "i'd": "i would",
        "i'll": "i will",
        "i'm": "i am",
        "i've": "i have",
        "i,d": "i would",
        "i,ll": "i will",
        "i,m": "i am",
        "i,ve": "i have",
        "i;d": "i would",
        "i;ll": "i will",
        "i;m": "i am",
        "i;ve": "i have",
        "isn't": "is not",
        "isn,t": "is not",
        "isn;t": "is not",
        "isn´t": "is not",
        "isn’t": "is not",
        "it'd": "it would",
        "it'll": "it will",
        "It's": "it is",
        "it's": "it is",
        "it,d": "it would",
        "it,ll": "it will",
        "it,s": "it is",
        "it;d": "it would",
        "it;ll": "it will",
        "it;s": "it is",
        "it´d": "it would",
        "it´ll": "it will",
        "it´s": "it is",
        "it’d": "it would",
        "it’ll": "it will",
        "it’s": "it is",
        "i´d": "i would",
        "i´ll": "i will",
        "i´m": "i am",
        "i´ve": "i have",
        "i’d": "i would",
        "i’ll": "i will",
        "i’m": "i am",
        "i’ve": "i have",
        "let's": "let us",
        "let,s": "let us",
        "let;s": "let us",
        "let´s": "let us",
        "let’s": "let us",
        "ma'am": "madam",
        "ma,am": "madam",
        "ma;am": "madam",
        "mayn't": "may not",
        "mayn,t": "may not",
        "mayn;t": "may not",
        "mayn´t": "may not",
        "mayn’t": "may not",
        "ma´am": "madam",
        "ma’am": "madam",
        "might've": "might have",
        "might,ve": "might have",
        "might;ve": "might have",
        "mightn't": "might not",
        "mightn,t": "might not",
        "mightn;t": "might not",
        "mightn´t": "might not",
        "mightn’t": "might not",
        "might´ve": "might have",
        "might’ve": "might have",
        "must've": "must have",
        "must,ve": "must have",
        "must;ve": "must have",
        "mustn't": "must not",
        "mustn,t": "must not",
        "mustn;t": "must not",
        "mustn´t": "must not",
        "mustn’t": "must not",
        "must´ve": "must have",
        "must’ve": "must have",
        "needn't": "need not",
        "needn,t": "need not",
        "needn;t": "need not",
        "needn´t": "need not",
        "needn’t": "need not",
        "oughtn't": "ought not",
        "oughtn,t": "ought not",
        "oughtn;t": "ought not",
        "oughtn´t": "ought not",
        "oughtn’t": "ought not",
        "sha'n't": "shall not",
        "sha,n,t": "shall not",
        "sha;n;t": "shall not",
        "shan't": "shall not",
        "shan,t": "shall not",
        "shan;t": "shall not",
        "shan´t": "shall not",
        "shan’t": "shall not",
        "sha´n´t": "shall not",
        "sha’n’t": "shall not",
        "she'd": "she would",
        "she'll": "she will",
        "she's": "she is",
        "she,d": "she would",
        "she,ll": "she will",
        "she,s": "she is",
        "she;d": "she would",
        "she;ll": "she will",
        "she;s": "she is",
        "she´d": "she would",
        "she´ll": "she will",
        "she´s": "she is",
        "she’d": "she would",
        "she’ll": "she will",
        "she’s": "she is",
        "should've": "should have",
        "should,ve": "should have",
        "should;ve": "should have",
        "shouldn't": "should not",
        "shouldn,t": "should not",
        "shouldn;t": "should not",
        "shouldn´t": "should not",
        "shouldn’t": "should not",
        "should´ve": "should have",
        "should’ve": "should have",
        "that'd": "that would",
        "that's": "that is",
        "that,d": "that would",
        "that,s": "that is",
        "that;d": "that would",
        "that;s": "that is",
        "that´d": "that would",
        "that´s": "that is",
        "that’d": "that would",
        "that’s": "that is",
        "there'd": "there had",
        "there's": "there is",
        "there,d": "there had",
        "there,s": "there is",
        "there;d": "there had",
        "there;s": "there is",
        "there´d": "there had",
        "there´s": "there is",
        "there’d": "there had",
        "there’s": "there is",
        "they'd": "they would",
        "they'll": "they will",
        "they're": "they are",
        "they've": "they have",
        "they,d": "they would",
        "they,ll": "they will",
        "they,re": "they are",
        "they,ve": "they have",
        "they;d": "they would",
        "they;ll": "they will",
        "they;re": "they are",
        "they;ve": "they have",
        "they´d": "they would",
        "they´ll": "they will",
        "they´re": "they are",
        "they´ve": "they have",
        "they’d": "they would",
        "they’ll": "they will",
        "they’re": "they are",
        "they’ve": "they have",
        "wasn't": "was not",
        "wasn,t": "was not",
        "wasn;t": "was not",
        "wasn´t": "was not",
        "wasn’t": "was not",
        "we'd": "we would",
        "we'll": "we will",
        "we're": "we are",
        "we've": "we have",
        "we,d": "we would",
        "we,ll": "we will",
        "we,re": "we are",
        "we,ve": "we have",
        "we;d": "we would",
        "we;ll": "we will",
        "we;re": "we are",
        "we;ve": "we have",
        "weren't": "were not",
        "weren,t": "were not",
        "weren;t": "were not",
        "weren´t": "were not",
        "weren’t": "were not",
        "we´d": "we would",
        "we´ll": "we will",
        "we´re": "we are",
        "we´ve": "we have",
        "we’d": "we would",
        "we’ll": "we will",
        "we’re": "we are",
        "we’ve": "we have",
        "what'll": "what will",
        "what're": "what are",
        "what's": "what is",
        "what've": "what have",
        "what,ll": "what will",
        "what,re": "what are",
        "what,s": "what is",
        "what,ve": "what have",
        "what;ll": "what will",
        "what;re": "what are",
        "what;s": "what is",
        "what;ve": "what have",
        "what´ll": "what will",
        "what´re": "what are",
        "what´s": "what is",
        "what´ve": "what have",
        "what’ll": "what will",
        "what’re": "what are",
        "what’s": "what is",
        "what’ve": "what have",
        "where'd": "where did",
        "where's": "where is",
        "where,d": "where did",
        "where,s": "where is",
        "where;d": "where did",
        "where;s": "where is",
        "where´d": "where did",
        "where´s": "where is",
        "where’d": "where did",
        "where’s": "where is",
        "who'll": "who will",
        "who's": "who is",
        "who,ll": "who will",
        "who,s": "who is",
        "who;ll": "who will",
        "who;s": "who is",
        "who´ll": "who will",
        "who´s": "who is",
        "who’ll": "who will",
        "who’s": "who is",
        "won't": "will not",
        "won,t": "will not",
        "won;t": "will not",
        "won´t": "will not",
        "won’t": "will not",
        "wouldn't": "would not",
        "wouldn,t": "would not",
        "wouldn;t": "would not",
        "wouldn´t": "would not",
        "wouldn’t": "would not",
        "you'd": "you would",
        "you'll": "you will",
        "you're": "you are",
        "you,d": "you would",
        "you,ll": "you will",
        "you,re": "you are",
        "you;d": "you would",
        "you;ll": "you will",
        "you;re": "you are",
        "you´d": "you would",
        "you´ll": "you will",
        "you´re": "you are",
        "you’d": "you would",
        "you’ll": "you will",
        "you’re": "you are",
        "´cause": "because",
        "’cause": "because",
        "you've": "you have",
        "could'nt": "could not",
        "havn't": "have not",
        "here’s": "here is",
        'i""m': "i am",
        "i'am": "i am",
        "i'l": "i will",
        "i'v": "i have",
        "wan't": "want",
        "was'nt": "was not",
        "who'd": "who would",
        "who're": "who are",
        "who've": "who have",
        "why'd": "why would",
        "would've": "would have",
        "y'all": "you all",
        "y'know": "you know",
        "you.i": "you i",
        "your'e": "you are",
        "arn't": "are not",
        "agains't": "against",
        "c'mon": "common",
        "doens't": "does not",
        'don""t': "do not",
        "dosen't": "does not",
        "dosn't": "does not",
        "shoudn't": "should not",
        "that'll": "that will",
        "there'll": "there will",
        "there're": "there are",
        "this'll": "this all",
        "u're": "you are",
        "ya'll": "you all",
        "you'r": "you are",
        "you’ve": "you have",
        "d'int": "did not",
        "did'nt": "did not",
        "din't": "did not",
        "dont't": "do not",
        "gov't": "government",
        "i'ma": "i am",
        "is'nt": "is not",
        "‘I": "I",
        "ᴀɴᴅ": "and",
        "ᴛʜᴇ": "the",
        "ʜᴏᴍᴇ": "home",
        "ᴜᴘ": "up",
        "ʙʏ": "by",
        "ᴀᴛ": "at",
        "…and": "and",
        "civilbeat": "civil beat",
        "TrumpCare": "Trump care",
        "Trumpcare": "Trump care",
        "OBAMAcare": "Obama care",
        "ᴄʜᴇᴄᴋ": "check",
        "ғᴏʀ": "for",
        "ᴛʜɪs": "this",
        "ᴄᴏᴍᴘᴜᴛᴇʀ": "computer",
        "ᴍᴏɴᴛʜ": "month",
        "ᴡᴏʀᴋɪɴɢ": "working",
        "ᴊᴏʙ": "job",
        "ғʀᴏᴍ": "from",
        "Sᴛᴀʀᴛ": "start",
        "gubmit": "submit",
        "CO₂": "carbon dioxide",
        "ғɪʀsᴛ": "first",
        "ᴇɴᴅ": "end",
        "ᴄᴀɴ": "can",
        "ʜᴀᴠᴇ": "have",
        "ᴛᴏ": "to",
        "ʟɪɴᴋ": "link",
        "ᴏғ": "of",
        "ʜᴏᴜʀʟʏ": "hourly",
        "ᴡᴇᴇᴋ": "week",
        "ᴇɴᴅ": "end",
        "ᴇxᴛʀᴀ": "extra",
        "Gʀᴇᴀᴛ": "great",
        "sᴛᴜᴅᴇɴᴛs": "student",
        "sᴛᴀʏ": "stay",
        "ᴍᴏᴍs": "mother",
        "ᴏʀ": "or",
        "ᴀɴʏᴏɴᴇ": "anyone",
        "ɴᴇᴇᴅɪɴɢ": "needing",
        "ᴀɴ": "an",
        "ɪɴᴄᴏᴍᴇ": "income",
        "ʀᴇʟɪᴀʙʟᴇ": "reliable",
        "ғɪʀsᴛ": "first",
        "ʏᴏᴜʀ": "your",
        "sɪɢɴɪɴɢ": "signing",
        "ʙᴏᴛᴛᴏᴍ": "bottom",
        "ғᴏʟʟᴏᴡɪɴɢ": "following",
        "Mᴀᴋᴇ": "make",
        "ᴄᴏɴɴᴇᴄᴛɪᴏɴ": "connection",
        "ɪɴᴛᴇʀɴᴇᴛ": "internet",
        "financialpost": "financial post",
        "ʜaᴠᴇ": " have ",
        "ᴄaɴ": " can ",
        "Maᴋᴇ": " make ",
        "ʀᴇʟɪaʙʟᴇ": " reliable ",
        "ɴᴇᴇᴅ": " need ",
        "ᴏɴʟʏ": " only ",
        "ᴇxᴛʀa": " extra ",
        "aɴ": " an ",
        "aɴʏᴏɴᴇ": " anyone ",
        "sᴛaʏ": " stay ",
        "Sᴛaʀᴛ": " start",
        "SHOPO": "shop",
    }
    punct = "/-'?!.,#$%'()*+-/:;<=>@[\\]^_`{|}~" + \
        '""“”’' + "∞θ÷α•à−β∅³π‘₹´°£€\×™√²—–&"
    punct_mapping = {
        "‘": "'",
        "₹": "e",
        "´": "'",
        "°": "",
        "€": "e",
        "™": "tm",
        "√": " sqrt ",
        "×": "x",
        "²": "2",
        "—": "-",
        "–": "-",
        "’": "'",
        "_": "-",
        "`": "'",
        "“": '"',
        "”": '"',
        "“": '"',
        "£": "e",
        "∞": "infinity",
        "θ": "theta",
        "÷": "/",
        "α": "alpha",
        "•": ".",
        "à": "a",
        "−": "-",
        "β": "beta",
        "∅": "",
        "³": "3",
        "π": "pi",
    }
    mispell_dict = {
        "SB91": "senate bill",
        "tRump": "trump",
        "utmterm": "utm term",
        "FakeNews": "fake news",
        "Gʀᴇat": "great",
        "ʙᴏᴛtoᴍ": "bottom",
        "washingtontimes": "washington times",
        "garycrum": "gary crum",
        "htmlutmterm": "html utm term",
        "RangerMC": "car",
        "TFWs": "tuition fee waiver",
        "SJWs": "social justice warrior",
        "Koncerned": "concerned",
        "Vinis": "vinys",
        "Yᴏᴜ": "you",
        "Trumpsters": "trump",
        "Trumpian": "trump",
        "bigly": "big league",
        "Trumpism": "trump",
        "Yoyou": "you",
        "Auwe": "wonder",
        "Drumpf": "trump",
        "utmterm": "utm term",
        "Brexit": "british exit",
        "utilitas": "utilities",
        "ᴀ": "a",
        "😉": "wink",
        "😂": "joy",
        "😀": "stuck out tongue",
        "theguardian": "the guardian",
        "deplorables": "deplorable",
        "theglobeandmail": "the globe and mail",
        "justiciaries": "justiciary",
        "creditdation": "Accreditation",
        "doctrne": "doctrine",
        "fentayal": "fentanyl",
        "designation-": "designation",
        "CONartist": "con-artist",
        "Mutilitated": "Mutilated",
        "Obumblers": "bumblers",
        "negotiatiations": "negotiations",
        "dood-": "dood",
        "irakis": "iraki",
        "cooerate": "cooperate",
        "COx": "cox",
        "racistcomments": "racist comments",
        "envirnmetalists": "environmentalists",
    }
    avg_glove_vector_840b300d = [
        0.22418134,
        -0.28881392,
        0.13854356,
        0.00365387,
        -0.12870757,
        0.10243822,
        0.061626635,
        0.07318011,
        -0.061350107,
        -1.3477012,
        0.42037755,
        -0.063593924,
        -0.09683349,
        0.18086134,
        0.23704372,
        0.014126852,
        0.170096,
        -1.1491593,
        0.31497982,
        0.06622181,
        0.024687296,
        0.076693475,
        0.13851812,
        0.021302193,
        -0.06640582,
        -0.010336159,
        0.13523154,
        -0.042144544,
        -0.11938788,
        0.006948221,
        0.13333307,
        -0.18276379,
        0.052385733,
        0.008943111,
        -0.23957317,
        0.08500333,
        -0.006894406,
        0.0015864656,
        0.063391194,
        0.19177166,
        -0.13113557,
        -0.11295479,
        -0.14276934,
        0.03413971,
        -0.034278486,
        -0.051366422,
        0.18891625,
        -0.16673574,
        -0.057783455,
        0.036823478,
        0.08078679,
        0.022949161,
        0.033298038,
        0.011784158,
        0.05643189,
        -0.042776518,
        0.011959623,
        0.011552498,
        -0.0007971594,
        0.11300405,
        -0.031369694,
        -0.0061559738,
        -0.009043574,
        -0.415336,
        -0.18870236,
        0.13708843,
        0.005911723,
        -0.113035575,
        -0.030096142,
        -0.23908928,
        -0.05354085,
        -0.044904727,
        -0.20228513,
        0.0065645403,
        -0.09578946,
        -0.07391877,
        -0.06487607,
        0.111740574,
        -0.048649278,
        -0.16565254,
        -0.052037314,
        -0.078968436,
        0.13684988,
        0.0757494,
        -0.006275573,
        0.28693774,
        0.52017444,
        -0.0877165,
        -0.33010918,
        -0.1359622,
        0.114895485,
        -0.09744406,
        0.06269521,
        0.12118575,
        -0.08026362,
        0.35256687,
        -0.060017522,
        -0.04889904,
        -0.06828978,
        0.088740796,
        0.003964443,
        -0.0766291,
        0.1263925,
        0.07809314,
        -0.023164088,
        -0.5680669,
        -0.037892066,
        -0.1350967,
        -0.11351585,
        -0.111434504,
        -0.0905027,
        0.25174105,
        -0.14841858,
        0.034635577,
        -0.07334565,
        0.06320108,
        -0.038343467,
        -0.05413284,
        0.042197507,
        -0.090380974,
        -0.070528865,
        -0.009174437,
        0.009069661,
        0.1405178,
        0.02958134,
        -0.036431845,
        -0.08625681,
        0.042951006,
        0.08230793,
        0.0903314,
        -0.12279937,
        -0.013899368,
        0.048119213,
        0.08678239,
        -0.14450377,
        -0.04424887,
        0.018319942,
        0.015026873,
        -0.100526,
        0.06021201,
        0.74059093,
        -0.0016333034,
        -0.24960588,
        -0.023739101,
        0.016396184,
        0.11928964,
        0.13950661,
        -0.031624354,
        -0.01645025,
        0.14079992,
        -0.0002824564,
        -0.08052984,
        -0.0021310581,
        -0.025350995,
        0.086938225,
        0.14308536,
        0.17146006,
        -0.13943303,
        0.048792403,
        0.09274929,
        -0.053167373,
        0.031103406,
        0.012354865,
        0.21057427,
        0.32618305,
        0.18015954,
        -0.15881181,
        0.15322933,
        -0.22558987,
        -0.04200665,
        0.0084689725,
        0.038156632,
        0.15188617,
        0.13274793,
        0.113756925,
        -0.095273495,
        -0.049490947,
        -0.10265804,
        -0.27064866,
        -0.034567792,
        -0.018810693,
        -0.0010360252,
        0.10340131,
        0.13883452,
        0.21131058,
        -0.01981019,
        0.1833468,
        -0.10751636,
        -0.03128868,
        0.02518242,
        0.23232952,
        0.042052146,
        0.11731903,
        -0.15506615,
        0.0063580726,
        -0.15429358,
        0.1511722,
        0.12745973,
        0.2576985,
        -0.25486213,
        -0.0709463,
        0.17983761,
        0.054027,
        -0.09884228,
        -0.24595179,
        -0.093028545,
        -0.028203879,
        0.094398156,
        0.09233813,
        0.029291354,
        0.13110267,
        0.15682974,
        -0.016919162,
        0.23927948,
        -0.1343307,
        -0.22422817,
        0.14634751,
        -0.064993896,
        0.4703685,
        -0.027190214,
        0.06224946,
        -0.091360025,
        0.21490277,
        -0.19562101,
        -0.10032754,
        -0.09056772,
        -0.06203493,
        -0.18876675,
        -0.10963594,
        -0.27734384,
        0.12616494,
        -0.02217992,
        -0.16058226,
        -0.080475815,
        0.026953284,
        0.110732645,
        0.014894041,
        0.09416802,
        0.14299914,
        -0.1594008,
        -0.066080004,
        -0.007995227,
        -0.11668856,
        -0.13081996,
        -0.09237365,
        0.14741232,
        0.09180138,
        0.081735,
        0.3211204,
        -0.0036552632,
        -0.047030564,
        -0.02311798,
        0.048961394,
        0.08669574,
        -0.06766279,
        -0.50028914,
        -0.048515294,
        0.14144728,
        -0.032994404,
        -0.11954345,
        -0.14929578,
        -0.2388355,
        -0.019883996,
        -0.15917352,
        -0.052084364,
        0.2801028,
        -0.0029121689,
        -0.054581646,
        -0.47385484,
        0.17112483,
        -0.12066923,
        -0.042173345,
        0.1395337,
        0.26115036,
        0.012869649,
        0.009291686,
        -0.0026459037,
        -0.075331464,
        0.017840583,
        -0.26869613,
        -0.21820338,
        -0.17084768,
        -0.1022808,
        -0.055290595,
        0.13513643,
        0.12362477,
        -0.10980586,
        0.13980341,
        -0.20233242,
        0.08813751,
        0.3849736,
        -0.10653763,
        -0.06199595,
        0.028849555,
        0.03230154,
        0.023856193,
        0.069950655,
        0.19310954,
        -0.077677034,
        -0.144811,
    ]
    avg_glove_vector_6b50d = [
        -0.12920076,
        -0.28866628,
        -0.01224866,
        -0.05676644,
        -0.20210965,
        -0.08389011,
        0.33359843,
        0.16045167,
        0.03867431,
        0.17833012,
        0.04696583,
        -0.00285802,
        0.29099807,
        0.04613704,
        -0.20923874,
        -0.06613114,
        -0.06822549,
        0.07665912,
        0.3134014,
        0.17848536,
        -0.1225775,
        -0.09916984,
        -0.07495987,
        0.06413227,
        0.14441176,
        0.60894334,
        0.17463093,
        0.05335403,
        -0.01273871,
        0.03474107,
        -0.8123879,
        -0.04688699,
        0.20193407,
        0.2031118,
        -0.03935686,
        0.06967544,
        -0.01553638,
        -0.03405238,
        -0.06528071,
        0.12250231,
        0.13991883,
        -0.17446303,
        -0.08011883,
        0.0849521,
        -0.01041659,
        -0.13705009,
        0.20127155,
        0.10069408,
        0.00653003,
        0.01685157,
    ]
    avg_fasttext_2m300d = [
        -1.79720020e-02,
        4.59612617e-02,
        2.44339478e-03,
        1.14453539e-01,
        -2.44256777e-02,
        2.95815698e-03,
        1.03025117e-01,
        -8.68739766e-02,
        -7.52062656e-02,
        3.33940391e-02,
        -2.02820195e-02,
        1.04620800e00,
        -8.15427891e-02,
        -2.37318223e-02,
        1.08383076e-02,
        1.23609346e-02,
        -7.79699414e-03,
        -2.23746934e-02,
        -7.49582109e-02,
        -3.71978281e-02,
        8.45225293e-03,
        1.36429609e-02,
        1.13141182e-02,
        1.16137305e-02,
        1.78923281e-02,
        2.93113906e-02,
        2.33791348e-02,
        4.45499180e-02,
        3.13387656e-02,
        -5.82022891e-02,
        1.28034033e-02,
        6.50548906e-02,
        -5.06797070e-02,
        7.72173047e-02,
        -2.29646797e-02,
        3.88520117e-02,
        1.71532629e-03,
        1.70338457e-02,
        5.56611836e-02,
        -1.01905477e-01,
        3.67041523e-02,
        5.74241641e-02,
        -1.77972734e-02,
        6.12034492e-02,
        -4.36416680e-02,
        6.36411289e-02,
        -8.27315547e-02,
        3.75490508e-02,
        -1.78396484e-02,
        -1.28496953e-01,
        6.64232891e-02,
        1.05493500e-01,
        9.04765500e-01,
        -7.41661865e-03,
        -4.32927813e-02,
        -2.54867422e-02,
        -8.73220410e-03,
        2.10639160e-02,
        3.87381523e-02,
        -3.23720020e-02,
        2.68976562e-02,
        9.85046641e-02,
        8.94975879e-03,
        7.13236797e-02,
        6.74019141e-02,
        -4.23013594e-02,
        -3.24311895e-02,
        -2.18652187e-02,
        1.22457446e-03,
        -1.69734023e-02,
        5.71789961e-02,
        1.75010547e-02,
        5.41422656e-02,
        -1.03576953e-01,
        2.51064453e-02,
        5.84627305e-02,
        -2.65568812e-01,
        6.46930391e-02,
        -1.34845484e-01,
        -1.91715762e-02,
        3.80878375e-01,
        2.69076562e-02,
        -3.52429258e-02,
        -1.69530273e-02,
        1.07793914e-01,
        -1.86758438e-02,
        -2.67136758e-02,
        -3.22055469e-02,
        5.86570469e-02,
        7.08526406e-02,
        -2.68723008e-02,
        1.92197484e-01,
        2.70301465e-02,
        -1.94913164e-02,
        4.47472500e-02,
        4.61226680e-02,
        -4.06695801e-03,
        9.14075156e-02,
        2.77345449e-02,
        4.14191602e-03,
        2.13715762e-02,
        -2.92259395e-02,
        -4.92994687e-02,
        -1.45244111e-02,
        8.72082344e-02,
        7.49558438e-02,
        2.61175366e-03,
        -6.05494961e-02,
        1.71665586e-02,
        1.59877451e-02,
        -1.20828984e-01,
        -7.45510400e-03,
        6.95515313e-02,
        -3.91321211e-02,
        -7.08964625e-01,
        2.03243242e-02,
        -8.41571250e-02,
        4.52808047e-02,
        -9.51013281e-02,
        3.80058750e-02,
        -2.27888809e-02,
        -1.18743838e-02,
        2.64994678e-03,
        5.20781016e-02,
        -5.60906328e-02,
        4.11016953e-02,
        2.95097583e-03,
        -5.99409297e-02,
        2.78802063e-01,
        -4.49910273e-02,
        1.25225615e-02,
        4.84114180e-02,
        -8.44276270e-03,
        6.06409375e-02,
        -6.56986563e-02,
        2.60665020e-02,
        6.63916563e-02,
        2.79978613e-02,
        1.53083525e-02,
        -8.38284750e-01,
        -9.80946797e-02,
        7.07481953e-02,
        -1.79018281e-02,
        -1.96852203e-01,
        -4.31327344e-02,
        9.00138281e-02,
        6.33505508e-02,
        8.09894727e-03,
        -5.04032266e-02,
        6.08266172e-02,
        -2.99341836e-02,
        3.85930625e-02,
        -6.72227813e-02,
        1.06172568e-02,
        3.76806562e-02,
        -2.82583340e-02,
        -7.97553162e-04,
        -4.15865273e-02,
        -4.67399883e-02,
        -1.20315713e-02,
        6.93710000e-01,
        -1.15665117e-01,
        7.59228271e-03,
        2.35386504e-02,
        -4.73535625e-02,
        -2.91041562e-02,
        6.39238232e-03,
        -2.03673437e-02,
        -7.01801953e-02,
        -6.97110547e-02,
        2.89453047e-02,
        -8.13781885e-03,
        -4.22333047e-02,
        2.23304082e-02,
        1.35199590e-02,
        1.04268930e-01,
        4.66349570e-02,
        -5.46767969e-02,
        1.17977109e-02,
        -2.35034980e-02,
        -6.54309766e-02,
        -6.37768789e-02,
        -3.06396738e-02,
        3.84648682e-03,
        1.32344512e-02,
        5.18146641e-02,
        2.33360156e-02,
        6.76717187e-02,
        4.41813789e-02,
        -9.02206797e-02,
        2.00402237e-04,
        1.68000547e-02,
        7.26895703e-02,
        5.82857695e-02,
        7.91773535e-03,
        7.33317109e-02,
        -8.23000625e-02,
        3.43291914e-02,
        2.13939355e-02,
        -8.97733582e-04,
        1.62947686e-02,
        -3.40215308e-03,
        -3.65187622e-03,
        1.13290557e-02,
        -2.23267852e-02,
        7.43288281e-02,
        2.72833008e-03,
        3.15662461e-02,
        -4.44898008e-02,
        1.59659607e-03,
        -9.52254785e-03,
        5.41178562e-01,
        -8.59935000e-02,
        -7.34877588e-03,
        -5.13457383e-02,
        3.51231367e-02,
        -5.37677578e-02,
        8.89518203e-02,
        -1.24034219e-02,
        -2.90076309e-02,
        -4.45301445e-02,
        -1.53866531e-01,
        1.51414297e-02,
        3.40245996e-03,
        4.36298945e-02,
        -4.71652852e-02,
        -2.37308716e-03,
        -6.84343689e-04,
        -1.08431689e-02,
        1.52163955e-02,
        -6.94941641e-02,
        5.52748096e-03,
        -4.93338906e-02,
        -5.39452246e-03,
        -3.05559512e-02,
        -7.84589563e-01,
        2.31532129e-02,
        3.92418633e-02,
        7.04897578e-02,
        3.05945742e-02,
        2.05342793e-02,
        -2.50711963e-03,
        -1.07908672e-02,
        -2.40677129e-02,
        8.24847656e-02,
        5.70925703e-02,
        -9.27704375e-02,
        -4.90966602e-03,
        -3.39621758e-02,
        -2.83528887e-02,
        -4.09463789e-02,
        3.32492944e-03,
        2.37715723e-02,
        -4.22930117e-02,
        4.90904443e-03,
        4.00825508e-02,
        5.94003555e-02,
        3.44932539e-02,
        7.00529297e-02,
        -2.54125977e-02,
        -1.90498359e-02,
        -2.21944668e-02,
        -7.29946875e-02,
        3.34532617e-02,
        4.51727422e-02,
        -3.81429956e-03,
        5.02195352e-02,
        -5.68943359e-03,
        5.14126758e-02,
        -5.58987656e-02,
        2.53255840e-02,
        -5.08024570e-02,
        2.10721797e-02,
        1.08110758e-01,
        -5.11616162e-03,
        -3.50598164e-02,
        2.98414863e-02,
        5.91751641e-02,
        9.42860840e-03,
        -4.89523594e-02,
        -2.19117832e-02,
        -1.22236461e-01,
        -2.76476836e-02,
        -1.73402070e-02,
        -9.76283750e-02,
        7.31549531e-02,
        -4.90989141e-02,
        1.08870596e-02,
        -2.87949980e-02,
        -4.20323789e-02,
        -2.64264844e-03,
        -3.77317969e-03,
        -3.13873340e-02,
        1.04166969e-01,
        8.90471562e-02,
        8.65275000e-02,
        -4.02050195e-02,
        6.80396250e-02,
        -2.37387500e-02,
        -1.00368828e-02,
    ]

    def __init__(self, prepare_emb=False):
        # check hist (log), you can know, most data is < 180 (99.1%), < 160 (95.2%), < 170 (97.4%), 140 (91.8%)
        self.MAX_LEN = MAX_LEN
        self.embedding_matrix = None
        self.embedding_index = None
        self.vocab = None
        self.INPUT_DATA_DIR = (
            "../input/jigsaw-unintended-bias-in-toxicity-classification/"
        )

        self.test_df = None
        self.train_df = None
        self.df = None
        self._contraction_handled = False
        self._special_chars_handled = False
        self._text_preprocessed = False
        self.tokenizer = None

        self.x_train = None
        self.x_test = None
        self.y_train = None
        self.y_aux_train = None

        self.test_df_id = (
            None  # only id series is needed for generating submission csv file
        )

        self.do_emb_matrix_preparation = prepare_emb

        # self.BIN_FOLDER="/proc/driver/nvidia/"
        self.BIN_FOLDER = "./"
        # test_df_id = None # for generating result, need id
        # self.E_M_FILE = self.BIN_FOLDER+"embedding.mat"
        # self.DATA_TRAIN_FILE = self.BIN_FOLDER+"emb_train_features.bin"
        # self.DATA_TEST_FILE = self.BIN_FOLDER+"emb_test_features.bin"

    @property
    def DATA_TEST_FILE(self):
        return self.BIN_FOLDER + "emb_test_features.bin"

    @property
    def DATA_TRAIN_FILE(self):
        return self.BIN_FOLDER + "emb_train_features.bin"

    @property
    def E_M_FILE(self):
        return self.BIN_FOLDER + "embedding.mat"

    def read_train_test_df(self, train_only=False):
        if self.train_df is None:
            try:
                self.train_df = pd.read_csv(self.INPUT_DATA_DIR + "train.csv")
            except FileNotFoundError:
                INPUT_DATA_DIR = "/home/pengyu/works/input/jigsaw-unintended-bias-in-toxicity-classification/"

                if os.path.isfile(INPUT_DATA_DIR + "train.csv"):
                    self.INPUT_DATA_DIR = INPUT_DATA_DIR

                INPUT_DATA_DIR = self.BIN_FOLDER

                if os.path.isfile(INPUT_DATA_DIR + "train.csv"):
                    self.INPUT_DATA_DIR = INPUT_DATA_DIR

                INPUT_DATA_DIR = "/content/gdrivedata/My Drive/"

                if os.path.isfile(INPUT_DATA_DIR + "train.csv"):
                    self.INPUT_DATA_DIR = INPUT_DATA_DIR

                self.train_df = pd.read_csv(self.INPUT_DATA_DIR + "train.csv")

        if not train_only:
            self.test_df = pd.read_csv(self.INPUT_DATA_DIR + "test.csv")
            self.test_df_id = (
                self.test_df.id
            )  # only id series is needed for generating submission csv file
            self.df = pd.concat(
                [self.train_df.iloc[:, [0, 2]], self.test_df.iloc[:, :2]]
            )

    def build_vocab(self, texts):
        if self.vocab:
            return  # no need to rebuild
        sentences = texts.apply(lambda x: x.split()).values  # for pandas data
        vocab = {}

        for sentence in sentences:
            for word in sentence:
                try:
                    vocab[word] += 1
                except KeyError:
                    vocab[word] = 1
        self.vocab = vocab

    def test_coverage(self):
        self.build_vocab(self.df[TEXT_COLUMN])  # test data text is known, too
        self.embedding_index = EmbeddingHandler.load_embeddings(
            EMBEDDING_FILES[0])
        EmbeddingHandler.check_coverage(self.vocab, self.embedding_index)

    def add_lower_to_embedding(self, embedding, vocab):
        count = 0

        for word in vocab:
            if word in embedding and word.lower() not in embedding:
                embedding[word.lower()] = embedding[word]
                count += 1
        utils.logger.debug(f"Added {count} words to embedding")

    def contraction_normalize(self):
        def clean_contractions(text, mapping):
            specials = ["’", "‘", "´", "`"]

            for s in specials:
                text = text.replace(s, "'")
            text = " ".join(
                [mapping[t] if t in mapping else t for t in text.split(" ")]
            )

            return text

        if not self._contraction_handled:
            if "lowered_comment" not in self.df.columns:
                self.df["lowered_comment"] = self.df[TEXT_COLUMN].apply(
                    lambda x: x.lower()
                )
            self.df["treated_comment"] = self.df["lowered_comment"].apply(
                lambda x: clean_contractions(
                    x, EmbeddingHandler.contraction_mapping)
            )
            self._contraction_handled = True
        # vocab = build_vocab(df['treated_comment']) # about 13.51 coverage

    def special_chars_normalize(self):
        def clean_special_chars(text, punct, mapping):
            for p in mapping:
                text = text.replace(p, mapping[p])

            for p in punct:
                text = text.replace(p, f" {p} ")
            specials = {
                "\u200b": " ",
                "…": " ... ",
                "\ufeff": "",
                "करना": "",
                "है": "",
            }  # Other special characters that I have to deal with in last

            for s in specials:
                text = text.replace(s, specials[s])

            return text

        if not self._special_chars_handled:
            if not self._contraction_handled:
                self.contraction_normalize()
            self.df["treated_comment"] = self.df["treated_comment"].apply(
                lambda x: clean_special_chars(
                    x, EmbeddingHandler.punct, EmbeddingHandler.punct_mapping
                )
            )
            self._special_chars_handled = True

    def spelling_normalize(self):
        def correct_spelling(x, dic):
            for word in dic.keys():
                x = x.replace(word, dic[word])

            return x

        if not self._special_chars_handled:
            self.special_chars_normalize()
        self.df["treated_comment"] = self.df["treated_comment"].apply(
            lambda x: correct_spelling(x, EmbeddingHandler.mispell_dict)
        )

    # vocab = build_vocab(df['treated_comment'])
    # print("Glove : ")
    # oov_glove = check_coverage(vocab, embed_glove)

    @staticmethod
    def check_coverage(vocab, embeddings_index):
        import operator

        known_words = {}
        unknown_words = {}
        nb_known_words = 0
        nb_unknown_words = 0

        for word in vocab.keys():
            try:
                known_words[word] = embeddings_index[word]
                nb_known_words += vocab[word]
            except KeyError:
                unknown_words[word] = vocab[word]
                nb_unknown_words += vocab[word]
                pass

        print(
            "Found embeddings for {:.2%} of vocab".format(
                len(known_words) / len(vocab))
        )
        print(
            "Found embeddings for  {:.2%} of all text".format(
                nb_known_words / (nb_known_words + nb_unknown_words)
            )
        )
        unknown_words = sorted(unknown_words.items(),
                               key=operator.itemgetter(1))[::-1]

        return unknown_words

    def get_identity_df(self):
        if (not self._text_preprocessed):
            # then we might be restore from numpy pickle file, so still need to read csv
            self.read_train_test_df()
            # for column in IDENTITY_COLUMNS :
            #    # it seems the .values will make a copy out, so it won't infect above sef.y_train
            #    self.train_df[column] = np.where(self.train_df[column] >= 0.5, True, False)
            #    # todo analyze >= 0.5 or not, what is the difference
            # refer to the paper, "This includes 450,000 comments annotated with the identities..."
        train_y_identity_df = (
            self.train_df[IDENTITY_COLUMNS]
            .dropna(how="all")
            .fillna(0)
            .astype(np.float32)
        )

        return train_y_identity_df

    def get_identity_train_data_df_idx(self):
        train_y_identity_df = self.get_identity_df()

        return (
            self.x_train[train_y_identity_df.index],
            train_y_identity_df.values,
            train_y_identity_df.index,
        )  # non-binary

    def _text_preprocess(self):
        self.spelling_normalize()  # will add 'treated_comment' column to df
        train_len = self.train_df.shape[0]
        self.train_df[TEXT_COLUMN] = self.df["treated_comment"][:train_len]
        self.test_df[TEXT_COLUMN] = self.df["treated_comment"][train_len:]

    def prepare_data(self, target_binarize=False):
        utils.logger.debug("Text preprocessing")

        if self._text_preprocessed:
            return

        self._text_preprocess()

        if target_binarize:
            for column in IDENTITY_COLUMNS + AUX_COLUMNS + [TARGET_COLUMN]:
                self.train_df[column] = np.where(
                    self.train_df[column] >= 0.5, True, False
                )

        # handle binary target data
        self.y_train = self.train_df[TARGET_COLUMN].values
        self.y_aux_train = self.train_df[AUX_COLUMNS].values

        x_train = self.train_df[TEXT_COLUMN].astype(str)
        x_test = self.test_df[TEXT_COLUMN].astype(str)

        # tokenizer = text.Tokenizer(filters=CHARS_TO_REMOVE)
        tokenizer = text.Tokenizer()
        tokenizer.fit_on_texts(list(x_train) + list(x_test))

        x_train = tokenizer.texts_to_sequences(x_train)
        x_test = tokenizer.texts_to_sequences(x_test)
        self.x_train = sequence.pad_sequences(
            x_train, maxlen=self.MAX_LEN
        )  # longer will be truncated
        self.x_test = sequence.pad_sequences(x_test, maxlen=self.MAX_LEN)

        self.tokenizer = tokenizer
        self._text_preprocessed = True
        utils.logger.debug("Text preprocessing Done")

    def _prepare_additional_data_pad(self, seqs):
        return sequence.pad_sequences(seqs, maxlen=self.MAX_LEN)

    def _prepare_additional_data(self):
        with open("wikisent2.txt") as f:
            lines = f.readlines()
            utils.dump_obj(self.tokenizer, "tokenizer")
            lines_seqs = self.tokenizer.texts_to_sequences(lines)
            utils.dump_obj(lines_seqs, "additional_data.bin", force=True)

    def build_matrix_prepare_data(
        self, path, emb_matrix_existed, convert_additional=False
    ):
        """
        build embedding matrix given tokenizer word_index and pre-trained embedding file

        :param word_index: word_index from tokenizer
        :param path: path to load pre-trained embedding
        :return: embedding matrix
        """

        if emb_matrix_existed:
            utils.logger.debug(
                "Start cooking embedding matrix and train/test data: only train/test data, emb_matrix existed"
            )
            self.prepare_data()

            if convert_additional:
                self._prepare_additional_data()

                return  # only need process text

        utils.logger.debug(
            "Start cooking embedding matrix and train/test data")
        utils.logger.debug(f"{path} is being processed")

        if path.find("840B.300d") > 0:
            emb_save_filename = "matrix_840b"

        if path.find("300d-2M") > 0:
            emb_save_filename = "matrix_crawl"

        emb_from_file = utils.get_obj_or_dump(emb_save_filename)

        if emb_from_file is not None:
            return emb_from_file

        if not utils.file_exist("word_index"):
            self.build_vocab(self.df[TEXT_COLUMN])
            vocab = utils.get_obj_or_dump(
                "vocab", self.vocab
            )  # word to integer value index
            self.prepare_data()  # tokenizer processed in this function, not related to embedding
            utils.logger.debug("Text processed")

            embedding_index = EmbeddingHandler.load_embeddings(
                path
            )  # embedding_index is an dict, value is the feature vector
            utils.logger.debug(f"loading embedding from {path} done")
            self.add_lower_to_embedding(
                embedding_index, vocab
            )  # will change embedding_index, add lower words in the vocab to this embedding

            word_index = utils.get_obj_or_dump(
                "word_index", self.tokenizer.word_index
            )  # word to integer value index
            embedding_matrix = np.zeros(
                (len(word_index) + 1, 300)
            )  # last one for unknown?
        else:
            utils.logger.debug("Restore word index from files")
            word_index = utils.get_obj_or_dump(
                "word_index"
            )  # word to integer value index
            try:
                vocab = utils.get_obj_or_dump(
                    "vocab", self.vocab
                )  # word to integer value index
            except FileNotFoundError:
                vocab = self.vocab

                if vocab is None:
                    raise RuntimeError(
                        "vocab shoule be None, process embedding_index need it"
                    )

            embedding_index = EmbeddingHandler.load_embeddings(
                path
            )  # embedding_index is an dict, value is the feature vector
            utils.logger.debug(f"loading embedding from {path} done")
            self.add_lower_to_embedding(
                embedding_index, vocab
            )  # will change embedding_index, add lower words in the vocab to this embedding

            embedding_matrix = np.zeros(
                (len(word_index) + 1, 300)
            )  # last one for unknown?

        if path.find("840B.300d") > 0:
            avg_vector = EmbeddingHandler.avg_glove_vector_840b300d

        if path.find("300d-2M") > 0:
            avg_vector = EmbeddingHandler.avg_fasttext_2m300d

        for word, i in word_index.items():
            try:
                embedding_matrix[i] = embedding_index[word]
            except KeyError:
                # for unk
                # https://stackoverflow.com/questions/49239941/what-is-unk-in-the-pretrained-glove-vector-files-e-g-glove-6b-50d-txt
                embedding_matrix[i] = avg_vector

        utils.logger.debug(
            f"Done cooking embedding matrix for {path} with train/test words"
        )
        utils.dump_obj(embedding_matrix, emb_save_filename, force=True)

        return embedding_matrix

    @staticmethod
    def get_coefs(word, *arr):
        # word embedding related

        return word, np.asarray(arr, dtype="float32")

    @staticmethod
    def load_embeddings(path):
        with open(path, "r") as f:
            return dict(
                EmbeddingHandler.get_coefs(*line.strip().split(" ")) for line in f
            )

    def prepare_tfrecord_data(
        self, dump=True, train_test_data=True, embedding=True, action=None
    ):
        if (
            action is not None and action == defaults.CONVERT_DATA_Y_NOT_BINARY
        ):  # unpicker, change y
            # only convert data y not binary
            self.read_train_test_df(train_only=True)

            if not os.path.isfile(DATA_FILE_FLAG):
                raise FileNotFoundError("Pickle files should be present")

            self.y_train = self.train_df[
                TARGET_COLUMN
            ].values  # not preprocessed, so still be float value

            if dump:
                train_data = zip(self.x_train, self.y_train, self.y_aux_train)
                pickle.dump(
                    train_data, open(self.DATA_TRAIN_FILE, "wb")
                )  # overwrite with new

                # file flag
                with open(DATA_FILE_FLAG, "wb") as f:
                    f.write(bytes("", "utf-8"))

            return (
                self.x_train,
                self.y_train,
                self.y_aux_train,
                self.x_test,
                self.embedding_matrix,
            )

        self.read_train_test_df()

        if (
            os.path.isfile(DATA_FILE_FLAG)
            and not self.do_emb_matrix_preparation
            and not train_test_data
            and not embedding
        ):
            # just recover from record file
            raise Exception("tfrecord should already existed, please check")

        if embedding:
            utils.logger.debug("Still need build embedding matrix")
            self.embedding_matrix = np.concatenate(
                [
                    self.build_matrix_prepare_data(f, emb_matrix_existed=False)

                    for f in EMBEDDING_FILES
                ],
                axis=-1,
            )
            del self.tokenizer

            if dump:
                if embedding:
                    pickle.dump(self.embedding_matrix,
                                open(self.E_M_FILE, "wb"))
            gc.collect()
            self.tokenizer = None
        elif (
            train_test_data
        ):  # no need to rebuild emb matrix, only need train test data
            f = EMBEDDING_FILES[0]

            if (
                action is not None
                and action == defaults.CONVERT_ADDITIONAL_NONTOXIC_DATA
            ):  # unpicker, change y
                self.build_matrix_prepare_data(
                    f, emb_matrix_existed=True, convert_additional=True
                )

                return
            self.build_matrix_prepare_data(f, emb_matrix_existed=True)

        # create an embedding_matrix
        # after this, embedding_matrix is a matrix of size
        # len(tokenizer.word_index)+1   x 600
        # we concatenate two matrices, 600 = 300+300 (crawl+glove)
        train_data = zip(
            self.x_train, self.y_train, self.y_aux_train
        )  # this need text_preprocess being called
        # global embedding_matrix

        if dump:
            if train_test_data:
                pickle.dump(train_data, open(self.DATA_TRAIN_FILE, "wb"))
                pickle.dump(self.x_test, open(self.DATA_TEST_FILE, "wb"))

            # file flag
            with open(DATA_FILE_FLAG, "wb") as f:
                f.write(bytes("", "utf-8"))

        return (
            self.x_train,
            self.y_train,
            self.y_aux_train,
            self.x_test,
            self.embedding_matrix,
        )

    def read_emb_data_from_input(self):
        emb_path = "../input/jigsaw-embedding-matrix/embedding.mat"
        data_path = "../input/jigsaw-vectors/emb_train_features.bin"
        test_data_path = "../input/jigsaw-vectors/emb_test_features.bin"
        emb = utils.get_obj_or_dump(emb_path, fullpath=True)
        data = utils.get_obj_or_dump(data_path, fullpath=True)
        test_data = utils.get_obj_or_dump(test_data_path, fullpath=True)

        return emb, data, test_data

    def data_prepare(self, action=None):
        """Returns the iris dataset as (train_x, train_y), (test_x, test_y).
        we load this from the tfrecord, maybe save the ones just after embedding, so it can be faster
        """

        if action is not None:
            utils.logger.debug("{} in data preparation".format(action))

        try:  # just recover from record file
            emb, data_train, test_data = self.read_emb_data_from_input()
            self.x_train, self.y_train, self.y_aux_train = zip(*data_train)
            self.x_train, self.y_train, self.y_aux_train = (
                np.array(self.x_train),
                np.array(self.y_train),
                np.array(self.y_aux_train),
            )
            utils.logger.debug("restored data from files for training")
            self.BIN_FOLDER = defaults.Data_Folder

            return self.x_train, self.y_train, self.y_aux_train, test_data, emb
        except FileNotFoundError:
            utils.logger.debug(
                "cannot restore emb, trainX from jigsaw kaggle file data"
            )
        except TypeError:  # if read out None for data_train
            utils.logger.debug(
                "cannot restore emb, trainX from jigsaw kaggle file data"
            )

        # if os.path.isfile(DATA_FILE_FLAG) and not self.do_emb_matrix_preparation:  # in final stage, no need to check this...

        if not self.do_emb_matrix_preparation:
            # global embedding_matrix

            if action is not None and action == defaults.DATA_ACTION_NO_NEED_LOAD_EMB_M:
                self.embedding_matrix = None
            else:
                try:
                    self.embedding_matrix = pickle.load(
                        open(self.E_M_FILE, "rb"))
                except FileNotFoundError:
                    self.BIN_FOLDER = "/content/gdrivedata/My Drive/"

                    if not os.path.isdir(self.BIN_FOLDER):
                        self.BIN_FOLDER = "./"

                        if not utils.file_exist(self.E_M_FILE, fullpath=True):
                            self.BIN_FOLDER = defaults.Data_Folder
                    self.embedding_matrix = pickle.load(
                        open(self.E_M_FILE, "rb"))
            utils.BIN_FOLDER = self.BIN_FOLDER  # save file to the right place
            utils.logger.debug(self.E_M_FILE)

            if action is not None:  # exist data, need to convert data
                utils.logger.debug(action)

                if (
                    action == defaults.CONVERT_TRAIN_DATA
                    or action == defaults.CONVERT_ADDITIONAL_NONTOXIC_DATA
                ):
                    self.prepare_tfrecord_data(
                        train_test_data=True, embedding=False, action=action
                    )  # train data will rebuild, so we put it before read from pickle

            try:
                data_train = pickle.load(
                    open(self.DATA_TRAIN_FILE, "rb")
                )  # (None, 2048)
            except FileNotFoundError:
                self.BIN_FOLDER = "/content/gdrivedata/My Drive/"

                if not os.path.isdir(self.BIN_FOLDER):
                    self.BIN_FOLDER = "./"

                utils.BIN_FOLDER = self.BIN_FOLDER  # save file to the right place
                data_train = pickle.load(
                    open(self.DATA_TRAIN_FILE, "rb")
                )  # (None, 2048)

            utils.logger.debug(self.DATA_TRAIN_FILE)
            self.x_test = pickle.load(
                open(self.DATA_TEST_FILE, "rb")
            )  # (None, 2048) 2048 features from xception net

            self.x_train, self.y_train, self.y_aux_train = zip(*data_train)
            self.x_train, self.y_train, self.y_aux_train = (
                np.array(self.x_train),
                np.array(self.y_train),
                np.array(self.y_aux_train),
            )

            # global test_df_id
            # test_df_id = pd.read_csv('../input/jigsaw-unintended-bias-in-toxicity-classification/test.csv').id

            try:
                self.test_df_id = pd.read_csv(
                    self.INPUT_DATA_DIR + "test.csv"
                ).id  # only id series is needed for generating submission csv file
            except FileNotFoundError:
                self.INPUT_DATA_DIR = "../input/"

                if not os.path.isdir(self.INPUT_DATA_DIR):
                    self.INPUT_DATA_DIR = "/home/pengyu/works/input/jigsaw-unintended-bias-in-toxicity-classification/"

                if not os.path.isdir(self.INPUT_DATA_DIR):
                    self.INPUT_DATA_DIR = (
                        self.BIN_FOLDER
                    )  # put same folder in google drive
                self.test_df_id = pd.read_csv(
                    self.INPUT_DATA_DIR + "test.csv"
                ).id  # only id series is needed for generating submission csv file

            if (
                action is not None
            ):  # exist data, need to convert data, so put after read from pickle
                if action == defaults.CONVERT_DATA_Y_NOT_BINARY:
                    self.prepare_tfrecord_data(
                        train_test_data=False, embedding=False, action=action
                    )  # train_test_data=False just not rebuild words, the y still need to change

            return (
                self.x_train,
                self.y_train,
                self.y_aux_train,
                self.x_test,
                self.embedding_matrix,
            )
        else:
            utils.logger.debug(self.DATA_TRAIN_FILE)
            # (x_train, y_train, y_aux_train), x_test = prepare_tfrecord_data()

            if action is not None and (
                action == defaults.CONVERT_TRAIN_DATA
                or action == defaults.CONVERT_ADDITIONAL_NONTOXIC_DATA
            ):
                self.embedding_matrix = pickle.load(open(self.E_M_FILE, "rb"))
                utils.logger.debug(
                    "Only build train test data, embedding loaded from pickle"
                )

                return self.prepare_tfrecord_data(embedding=False, action=action)
            else:
                return self.prepare_tfrecord_data(embedding=True)



import numpy as np
import torch

import torch.nn as nn
import sklearn
from kaggle_runner.datasets.coders import compute_iou_batch
from kaggle_runner.metrics.metrics import metric
from kaggle_runner.predictors import predict
from .metrics import matthews_correlation


class Meter:
    """A meter to keep track of iou and dice scores throughout an epoch"""

    def __init__(self, phase, epoch):
        self.base_threshold = 0.5  # <<<<<<<<<<< here's the threshold
        self.base_dice_scores = []
        self.dice_neg_scores = []
        self.dice_pos_scores = []
        self.iou_scores = []

    def update(self, targets, outputs):
        probs = torch.sigmoid(outputs)
        dice, dice_neg, dice_pos, _, _ = metric(probs, targets, self.base_threshold)
        self.base_dice_scores.append(dice)
        self.dice_pos_scores.append(dice_pos)
        self.dice_neg_scores.append(dice_neg)
        preds = predict(probs, self.base_threshold)
        iou = compute_iou_batch(preds, targets, classes=[1])
        self.iou_scores.append(iou)

    def get_metrics(self):
        dice = np.mean(self.base_dice_scores)
        dice_neg = np.mean(self.dice_neg_scores)
        dice_pos = np.mean(self.dice_pos_scores)
        dices = [dice, dice_neg, dice_pos]
        iou = np.nanmean(self.iou_scores)

        return dices, iou

# # + colab={} colab_type="code" id="swiDMY2l3bHb"
class RocAucMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.y_true = np.array([])
        self.y_true_float = np.array([], dtype=np.float)
        self.y_pred = np.array([])
        self.score = 0
        self.mc_score = 0
        self.aux_part = 0

    def update(self, y_true, y_pred, aux_part=0):
        #y_true_ = y_true
        y_true = y_true[:,:2].cpu().numpy().argmax(axis=1)
        y_true_float = y_true.astype(np.float)
        y_pred = nn.functional.softmax(y_pred[:,:2], dim=1).data.cpu().numpy()[:,1]
        self.y_true = np.hstack((self.y_true, y_true))
        self.y_true_float = np.hstack((self.y_true_float, y_true_float))
        self.y_pred = np.hstack((self.y_pred, y_pred))
        try:
            self.score = sklearn.metrics.roc_auc_score(self.y_true, self.y_pred, labels=np.array([0, 1]))
        except Exception:
            self.score = 0
        self.mc_score = matthews_correlation(self.y_true_float, self.y_pred)
        self.aux_part = aux_part

    @property
    def avg(self):
        return self.score
    @property
    def mc_avg(self):
        return self.mc_score

# # + colab={} colab_type="code" id="sj4NmKFm3bHj"
class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



from kaggle_runner.kernels import bert_torch
from kaggle_runner.datasets.bert import pack_data

from kaggle_runner import logger, may_debug

class Test_bert_multi_lang:
    @classmethod
    def setup_class(cls):
        try:
            cls.model = bert_torch.get_trained_model()
        except RuntimeError:
            cls.model = None
        cls.data=pack_data()
        logger.debug("Start Test bert multi lang")

    def setup_method(self, method):
        logger.debug("setup for method %s", method)

    def teardown_method(self, method):
        logger.debug("teardown method %s", method)

    def test_continue_train(self):
        bert_torch.for_pytorch(self.data, phase="continue_train", model=self.model)

    def test_result(self):
        bert_torch.get_test_result(self, self.data[-1])

if __name__ == "__main__":
    Test_bert_multi_lang.setup_class()
    t = Test_bert_multi_lang()
    t.test_continue_train()
    t.test_result()



import math

import numpy as np
import torch
from torch.optim.optimizer import Optimizer, required


class RAdam(Optimizer):
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        self.buffer = [[None, None, None] for ind in range(10)]
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state["step"] += 1
                buffered = self.buffer[int(state["step"] % 10)]
                if state["step"] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state["step"]
                    beta2_t = beta2 ** state["step"]
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = (
                            group["lr"]
                            * math.sqrt(
                                (1 - beta2_t)
                                * (N_sma - 4)
                                / (N_sma_max - 4)
                                * (N_sma - 2)
                                / N_sma
                                * N_sma_max
                                / (N_sma_max - 2)
                            )
                            / (1 - beta1 ** state["step"])
                        )
                    else:
                        step_size = group["lr"] / (1 - beta1 ** state["step"])
                    buffered[2] = step_size

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(-step_size, exp_avg, denom)
                else:
                    p_data_fp32.add_(-step_size, exp_avg)

                p.data.copy_(p_data_fp32)

        return loss


class EarlyStopping(object):
    """EarlyStop for pytorch
    refer to
    https://gist.github.com/stefanonardo/693d96ceb2f531fa05db530f3e21517d

    TODO check if fastai has buildin support for this
    """

    def __init__(self, mode="min", min_delta=0, patience=10, percentage=False):
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta, percentage)

        if patience == 0:
            self.is_better = lambda a, b: True
            self.step = lambda a: False

    def step(self, metrics):
        if self.best is None:
            self.best = metrics
            return False

        if np.isnan(metrics):
            return True

        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.best = metrics
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            return True

        return False

    def _init_is_better(self, mode, min_delta, percentage):
        if mode not in {"min", "max"}:
            raise ValueError("mode " + mode + " is unknown!")
        if not percentage:
            if mode == "min":
                self.is_better = lambda a, best: a < best - min_delta
            if mode == "max":
                self.is_better = lambda a, best: a > best + min_delta
        else:
            if mode == "min":
                self.is_better = lambda a, best: a < best - (best * min_delta / 100)
            if mode == "max":
                self.is_better = lambda a, best: a > best + (best * min_delta / 100)



import os
import pickle
import re

import numpy as np
import pandas as pd
import tensorflow as tf

from kaggle_runner import may_debug, logger
from kaggle_runner.defaults import DEBUG, LOAD_BERT_DATA
from kaggle_runner.utils.kernel_utils import (get_kaggle_dataset_input,
                                              get_obj_or_dump)
#from kaggle_runner.utils.tpu import (strategy, tpu_resolver)
from hub.bert.extract_features import load_data, get_tokenizer
import torch
import torch.nn as nn
from torch.utils.data import Dataset,DataLoader
from torch.autograd import Variable
from torch.utils.data.sampler import SequentialSampler, RandomSampler
from fastai import *
from fastai.core import *


TOKENS_PATH = "/kaggle/input/jigsaw-toxic-token-ids-for-bert"
PRETRAIND_PICKLE_AND_MORE='/kaggle/input/toxic-multilang-trained-torch-model'
TRAIN_LEN = 0

strategy=None
tpu_resolver=None
BATCH_SIZE = 32 * 2
MAX_LENGTH = 224
X,y, x_valid, y_valid, X_test = None,None,None,None,None

if tpu_resolver is None:
    if DEBUG:
        BATCH_SIZE = 32 * 2
    else:
        BATCH_SIZE = 32 * 32
elif strategy is not None:
    BATCH_SIZE = 32 * strategy.num_replicas_in_sync

if tpu_resolver is None:
    DATA_PATH = "/kaggle/input/jigsaw-multilingual-toxic-comment-classification/"
    BERT_BASE_DIR = "/kaggle/input/bert-pretrained-models" + \
        '/multi_cased_L-12_H-768_A-12' + '/multi_cased_L-12_H-768_A-12'
else:
    from kaggle_datasets import KaggleDatasets
    GCS_DS_PATH = KaggleDatasets().get_gcs_path(
        'jigsaw-multilingual-toxic-comment-classification')
    GCS_BERT_PRETRAINED = KaggleDatasets().get_gcs_path('bert-pretrained-models') + \
        '/multi_cased_L-12_H-768_A-12'+'/multi_cased_L-12_H-768_A-12'

    DATA_PATH = GCS_DS_PATH + '/'
    BERT_BASE_DIR = GCS_BERT_PRETRAINED


def pickle_data(max_seq_length=128, bert_base_dir=BERT_BASE_DIR, output="features.pkl"):
    # --vocab_file="$BERT_BASE_DIR/vocab.txt" \
    # --init_checkpoint="$BERT_BASE_DIR/bert_model.ckpt" \
    # --bert_config_file="$BERT_BASE_DIR/bert_config.json" \
    load_data("pickle", "/tmp/input.txt", max_seq_length,
              get_tokenizer(bert_base_dir+"/vocab.txt"), output=output)


def load_tokens(data_base_dir=TOKENS_PATH, max_seq_length=128, bert_base_dir=BERT_BASE_DIR, output=None):
    tks = load_data("load_tokens", f'{data_base_dir}/token_ids_{max_seq_length}.pkl',
                    max_seq_length, get_tokenizer(bert_base_dir+"/vocab.txt"), output=output)

    return tks



#if DEBUG:
tf.executing_eagerly()
# Dataloading related
AUTO = tf.data.experimental.AUTOTUNE

# ### Load the training, validation, and testing datasets


# +
TEST_PATH = DATA_PATH + "test.csv"
VAL_PATH = DATA_PATH + "validation.csv"
TRAIN_PATH = DATA_PATH + "jigsaw-toxic-comment-train.csv"

val_data = None
test_data = None
train_data = None

# +
train_dataset = None
test_dataset = None
valid_dataset = None

if LOAD_BERT_DATA:
    data_package = get_kaggle_dataset_input(
        "jigsaw-multilingula-toxicity-token-encoded/toxic_fast_tok_512.pk")
    try:
        csv_data_package = get_obj_or_dump("toxic_csv.pk")

        if csv_data_package is None:
            csv_data_package = get_kaggle_dataset_input(
                "jigsaw-multilingula-toxicity-token-encoded/toxic_csv.pk")
    except ModuleNotFoundError as e:
        logger.error("%s", e)
        csv_data_package = None

    if csv_data_package is None:
        val_data = pd.read_csv(VAL_PATH)
        test_data = pd.read_csv(TEST_PATH)
        train_data = pd.read_csv(TRAIN_PATH)
        csv_data_package = (val_data, test_data, train_data)
        get_obj_or_dump("toxic_csv.pk", default=csv_data_package)
    else:
        val_data, test_data, train_data = csv_data_package


    if data_package is None:
        import transformers
        from tokenizers import BertWordPieceTokenizer
        tokenizer = transformers.DistilBertTokenizer.from_pretrained(
            'distilbert-base-multilingual-cased')

        save_path = '/kaggle/working/distilbert_base_uncased/'

        if not os.path.exists(save_path):
            os.makedirs(save_path)
        tokenizer.save_pretrained(save_path)

        def fast_encode(texts, tokenizer, chunk_size=240, maxlen=512):
            tokenizer.enable_truncation(max_length=maxlen)
            tokenizer.enable_padding(max_length=maxlen)
            all_ids = []

            for i in range(0, len(texts), chunk_size):
                text_chunk = texts[i:i+chunk_size].tolist()
                encs = tokenizer.encode_batch(text_chunk)
                all_ids.extend([enc.ids for enc in encs])

            return np.array(all_ids)

        fast_tokenizer = BertWordPieceTokenizer('distilbert_base_uncased/vocab.txt',
                                                lowercase=True)

# ### Clean the text (remove usernames and links)
# +
        val = val_data
        train = train_data

        def clean(text):
            text = text.fillna("fillna").str.lower()
            text = text.map(lambda x: re.sub('\\n', ' ', str(x)))
            text = text.map(lambda x: re.sub(
                "\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", '', str(x)))
            text = text.map(lambda x: re.sub(
                "\(http://.*?\s\(http://.*\)", '', str(x)))

            return text

        val["comment_text"] = clean(val["comment_text"])
        test_data["content"] = clean(test_data["content"])
        train["comment_text"] = clean(train["comment_text"])


# -

        x_train = fast_encode(train.comment_text.astype(str),
                              fast_tokenizer, maxlen=512)
        x_valid = fast_encode(val_data.comment_text.astype(str).values,
                              fast_tokenizer, maxlen=512)
        x_test = fast_encode(test_data.content.astype(str).values,
                             fast_tokenizer, maxlen=512)

# TODO just save it to disk or dataset for faster startup, and use it as a
# dataset
        y_valid = val.toxic.values  # no aux info for validation
# y_train = train.toxic.values  # TODO add aux data
### Define training, validation, and testing datasets

        y_train = np.stack([train.toxic.values, train.severe_toxic.values,
                            train.obscene.values, train.threat.values,
                            train.insult.values, train.identity_hate.values]).T

        data_package = get_obj_or_dump("toxic_fast_tok_512.pk", default=(x_train,
                                                                         y_train,
                                                                         x_valid,
                                                                         y_valid,
                                                                         x_test))
    else:
        x_train, y_train, x_valid, y_valid, x_test = data_package

    TRAIN_LEN = len(x_train)
    VALID_LEN = len(x_valid)
    TEST_LEN = len(x_test)

    if DEBUG:
        TRAIN_LEN = TRAIN_LEN//10
        VALID_LEN = VALID_LEN//10
        x_train = x_train[:TRAIN_LEN, :140]
        y_train = y_train[:TRAIN_LEN]  # y just all pass, they are labels
        x_valid = x_valid[:VALID_LEN, :140]
        y_valid = y_valid[:VALID_LEN]  # it just one dimention
        # TEST_LEN = TEST_LEN//100
        # x_test  = x_test[:TEST_LEN, :140]
    train_dataset = (
        tf.data.Dataset
        .from_tensor_slices((x_train, y_train))
        .repeat()
        .shuffle(2048)
        .batch(BATCH_SIZE)
        .prefetch(AUTO)
    )

    valid_dataset = (
        tf.data.Dataset
        .from_tensor_slices((x_valid, y_valid))
        .shuffle(1024)
        .batch(BATCH_SIZE)
        .cache()
        .prefetch(AUTO)
    )

    test_dataset = (
        tf.data.Dataset
        .from_tensor_slices(x_test)
        .batch(BATCH_SIZE)
    )

def xlmr_data():
    xlmr = torch.hub.load('pytorch/fairseq', 'xlmr.large')
    xlmr.eval()

    xlmr.encode('Hello world!')
    del xlmr

def load_labels():
    return (y_train, y_valid)

def pack_data():
    tokens = load_tokens()
    lbs = load_labels()

    y = lbs[0]
    X = [ x.input_ids for x in tokens[:len(y)] ]

    y_val = lbs[1]
    X_val = [ x.input_ids for x in tokens[len(y):len(y)+len(y_val)]]

    X_test = [ x.input_ids for x in tokens[len(y)+len(y_val):]]

    if DEBUG:
        y = y[:100]
        X = X[:100]
        y_val = y_val[:100]
        X_val = X_val[:100]


    return X,y, X_val, y_val, X_test

def onehot(size, target, aux=None):
    if aux is not None:
        vec = np.zeros(size+len(aux), dtype=np.float32)
        vec[target] = 1.
        vec[2:] = aux
        vec = torch.tensor(vec, dtype=torch.float32)
    else:
        vec = torch.zeros(size, dtype=torch.float32)
        vec[target] = 1.

    return vec

class DatasetRetriever(Dataset):
    def __init__(self, labels_or_ids, comment_texts, langs,
                 severe_toxic=None, obscene=None, threat=None, insult=None, identity_hate=None,
                 use_train_transforms=False, test=False, use_aux=True, transformers=None):
        self.test = test
        self.labels_or_ids = labels_or_ids
        self.comment_texts = comment_texts
        self.langs = langs
        self.severe_toxic = severe_toxic
        self.obscene = obscene
        self.threat = threat
        self.insult = insult
        self.identity_hate = identity_hate
        self.use_train_transforms = use_train_transforms
        self.aux = None
        assert transformers is not None
        self.transformers = transformers
        self.vocab = None

        if use_aux:
            self.aux = [self.severe_toxic, self.obscene, self.threat, self.insult, self.identity_hate]

    def get_tokens(self, text):
        encoded = self.transformers['tokenizer'].encode_plus(
            text,
            add_special_tokens=True,
            max_length=MAX_LENGTH,
            pad_to_max_length=True
        )

        return encoded['input_ids'], encoded['attention_mask']

    def __len__(self):
        return self.comment_texts.shape[0]

    def __getitem__(self, idx):
        text = self.comment_texts[idx]
        lang = self.langs[idx]

        if self.severe_toxic is None:
            aux = [0., 0., 0., 0., 0.]
        else:
            aux = [self.severe_toxic[idx], self.obscene[idx], self.threat[idx], self.insult[idx], self.identity_hate[idx]]


        label = self.labels_or_ids[idx]

        if self.use_train_transforms and (not self.test):
            text, _ = self.transformers['train_transforms'](data=(text, lang))['data']
            tokens, attention_mask = self.get_tokens(str(text))
            token_length = sum(attention_mask)

            if token_length > 0.8*MAX_LENGTH:
                text, _ = self.transformers['shuffle_transforms'](data=(text, lang))['data']
            elif token_length < 60:
                text, label = self.transformers['synthesic_transforms_often'](data=(text, label))['data']
            else: # will not need to use transforms
                #text, label = synthesic_transforms_low(data=(text, label))['data']
                pass

        # TODO add language detection and shuffle
        # https://pypi.org/project/langdetect/
        # if self.use_train_transforms and self.test:
        #    text, _ = train_transforms(data=(text, lang))['data']
        #    tokens, attention_mask = self.get_tokens(str(text))
        #    token_length = sum(attention_mask)

        #    if token_length > 0.8*MAX_LENGTH:
        #        text, _ = shuffle_transforms(data=(text, lang))['data']
        # to tensors
        tokens, attention_mask = self.get_tokens(str(text))
        tokens, attention_mask = torch.tensor(tokens), torch.tensor(attention_mask)

        if self.test:  # for test, return id TODO TTA
            return [tokens, attention_mask], self.labels_or_ids[idx]

        # label might be changed
        target = onehot(2, label, aux=aux)

        return [tokens, attention_mask], target

    def get_labels(self):
        return list(np.char.add(self.labels_or_ids.astype(str), self.langs))



from __future__ import absolute_import, division, print_function
import sys
package_dir_a = "/kaggle/input/ppbert/pytorch-pretrained-bert/pytorch-pretrained-BERT"
sys.path.insert(0, package_dir_a)

from kaggle_runner.datasets.bert import DATA_PATH, BERT_BASE_DIR, PRETRAIND_PICKLE_AND_MORE
from kaggle_runner import may_debug, logger
from tqdm import tqdm, tqdm_notebook
from sklearn.metrics import roc_auc_score
from sklearn import metrics, model_selection
from pytorch_pretrained_bert import (BertAdam, BertForSequenceClassification,
                                     BertTokenizer, BertConfig,
                                     convert_tf_checkpoint_to_pytorch)  # needed fused_layer_norm_cuda, so TPU won't work

import datetime
import gc
import operator
import os
import pickle
import re
import shutil
import time
import warnings

import numpy as np  # linear algebra
import pandas as pd  # data processing, CSV file I/O (e.g. pd.read_csv)
import pkg_resources
import scipy.stats as stats
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data
from IPython.core.interactiveshell import InteractiveShell
from nltk.stem import PorterStemmer


# %load_ext autoreload
# %autoreload 2
# %matplotlib inline


InteractiveShell.ast_node_interactivity = "all"
warnings.filterwarnings(action='once')


EPOCHS = 3
WORK_DIR = '/kaggle/working/'


def prepare_pretrained():
    convert_tf_checkpoint_to_pytorch.convert_tf_checkpoint_to_pytorch(
        BERT_BASE_DIR + '/bert_model.ckpt',
        BERT_BASE_DIR + '/bert_config.json',
        WORK_DIR + 'pytorch_model.bin')

    shutil.copyfile(BERT_BASE_DIR + '/bert_config.json',
                    WORK_DIR + 'bert_config.json')
# -


def get_trained_model(fine_tuned = "bert_pytorch.bin", device=torch.device('cuda')):
    model = None
    y_columns = ['toxic', "severe_toxic","obscene","threat","insult","identity_hate"]
    pretrain_data_folder = PRETRAIND_PICKLE_AND_MORE

    if not os.path.exists(pretrain_data_folder+"/" + fine_tuned):
        pretrain_data_folder = '/home/working'

    if os.path.exists(pretrain_data_folder+"/"+fine_tuned):
        output_model_file = pretrain_data_folder+"/"+fine_tuned
        bert_config = BertConfig.from_json_file(pretrain_data_folder + "/bert_config.json")

        # Run validation
        # The following 2 lines are not needed but show how to download the model for prediction
        model = BertForSequenceClassification(
            bert_config, num_labels=len(y_columns))
        model.load_state_dict(torch.load(output_model_file))
        model.to(device)

    return model

def get_predict(model, test_dataset, device=torch.device('cuda')):
    if model is None:
        return

    for param in model.parameters():
        param.requires_grad = False
    model.eval()

    test_preds = np.zeros((len(test_dataset)))
    test = torch.utils.data.TensorDataset(
        torch.tensor(test_dataset, dtype=torch.long))
    test_loader = torch.utils.data.DataLoader(
        test, batch_size=32, shuffle=False)

    tk0 = tqdm(test_loader)

    for i, (x_batch,) in enumerate(tk0):
        pred = model(x_batch.to(device), attention_mask=(
            x_batch > 0).to(device), labels=None)
        test_preds[i*32:(i+1)*32] = pred[:, 0].detach().cpu().squeeze().numpy()

    return test_preds

def get_validation_result(model, test_dataset, y_targets, device=torch.device('cuda')) :
    test_preds = get_predict(model, test_dataset, device)
    pred = 1 / (1+np.exp(- test_preds))

    if len(y_targets.shape) == 1:
        return roc_auc_score(y_targets, pred)

    return roc_auc_score(y_targets[:,0], pred)

def get_test_result(self, test_dataset, device=torch.device('cuda'), data_path=DATA_PATH):
    test_preds = get_predict(self.model, test_dataset, device)
    pred = 1 / (1+np.exp(- test_preds))

    sub = pd.read_csv(os.path.join(data_path , 'sample_submission.csv'))
    sub['toxic'][:len(pred)] = pred
    sub.to_csv('submission.csv', index=False)


def for_pytorch(data_package, device=torch.device('cuda'), SEED=118, phase="predict", model=None):

    if device is None and os.getenv("TPU_NAME") is not None:
        import torch_xla # model
        import torch_xla.core.xla_model as xm
        device = xm.xla_device()

    X, y, X_val, y_val, X_test = data_package

    if model is None:
        try:
            model = get_trained_model(device=device)
        except RuntimeError as e:
            logger.debug("%s", e)

    if model is not None and phase=="predict":
        for param in model.parameters():
            param.requires_grad = False
        model.eval()
        valid_preds = np.zeros((len(X_val)))
        valid = torch.utils.data.TensorDataset(
            torch.tensor(X_val, dtype=torch.long))
        valid_loader = torch.utils.data.DataLoader(
            valid, batch_size=32, shuffle=False)

        tk0 = tqdm(valid_loader)

        for i, (x_batch,) in enumerate(tk0):
            pred = model(x_batch.to(device), attention_mask=(
                x_batch > 0).to(device), labels=None)
            valid_preds[i*32:(i+1)*32] = pred[:, 0].detach().cpu().squeeze().numpy()
    else:
        import subprocess
        train_dataset = torch.utils.data.TensorDataset(torch.tensor(
            X, dtype=torch.long), torch.tensor(y, dtype=torch.float))
        output_model_file = "bert_pytorch.bin"

        lr = 1e-5
        batch_size = 32
        accumulation_steps = 3
        np.random.seed(SEED)
        torch.manual_seed(SEED)
        torch.cuda.manual_seed(SEED)
        torch.backends.cudnn.deterministic = False

        if model is None:
            prepare_pretrained()
            model = BertForSequenceClassification.from_pretrained(
                ".", cache_dir=None, num_labels=1 if len(y[0]) < 1 else len(y[0]))
            assert model is not None
        logger.info("AUC for valication: %f", get_validation_result(model, X_val, y_val))
        model.zero_grad()
        model = model.to(device)

        param_optimizer = list(model.named_parameters())
        may_debug()

        req_grad = ['layer.10', 'layer.11', 'bert.poole', 'classifier']
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']

        def para_opt_configure(req_grad, no_decay):
            for n, p in param_optimizer:
                if any(nd in n for nd in req_grad):
                    p.requires_grad = True
                else:
                    p.requires_grad = False

            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(
                    nd in n for nd in no_decay)], 'weight_decay': 0.01},
                {'params': [p for n, p in param_optimizer if any(
                    nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]

            return optimizer_grouped_parameters
        optimizer_grouped_parameters = para_opt_configure(req_grad, no_decay)
        train = train_dataset

        num_train_optimization_steps = int(
            EPOCHS*len(train)/batch_size/accumulation_steps)

        optimizer = BertAdam(optimizer_grouped_parameters,
                       lr=lr,
                       warmup=0.05,
                       t_total=num_train_optimization_steps)

        subprocess.run('python3 -m pip show apex || ([ -d /kaggle/input/nvidiaapex/repository/NVIDIA-apex-39e153a ] && '
            'pip install -v --no-cache-dir --global-option="--cpp_ext" --global-option="--cuda_ext" ../input/nvidiaapex/repository/NVIDIA-apex-39e153a)',
                       shell=True, check=True)
        from apex import amp  # automatic mix precision
        model, optimizer = amp.initialize(
            model, optimizer, opt_level="O1", verbosity=1)
        model = model.train()

        tq = tqdm(range(EPOCHS))

        for epoch in tq:
            train_loader = torch.utils.data.DataLoader(
                train, batch_size=batch_size, shuffle=True)
            avg_loss = 0.
            avg_accuracy = 0.
            lossf = None
            para_opt_configure(req_grad, no_decay)  # valication will change it
            tk0 = tqdm(enumerate(train_loader), total=len(train_loader), leave=True)
            optimizer.zero_grad()   # Bug fix - thanks to @chinhuic

            for i, (x_batch, y_batch) in tk0:
                #        optimizer.zero_grad()
                y_pred = model(x_batch.to(device), attention_mask=(
                    x_batch > 0).to(device), labels=None)
                loss = F.binary_cross_entropy_with_logits(
                    y_pred, y_batch.to(device))
                with amp.scale_loss(loss, optimizer) as scaled_loss:
                    scaled_loss.backward()

                if (i+1) % accumulation_steps == 0:             # Wait for several backward steps
                    optimizer.step()                            # Now we can do an optimizer step
                    optimizer.zero_grad()

                if lossf:
                    lossf = 0.98*lossf+0.02*loss.item()
                else:
                    lossf = loss.item()
                tk0.set_postfix(loss=lossf)
                avg_loss += loss.item() / len(train_loader)
                avg_accuracy += torch.mean(((torch.sigmoid(y_pred[:, 0]) > 0.5) == (
                    y_batch[:, 0] > 0.5).to(device)).to(torch.float)).item()/len(train_loader)
            tq.set_postfix(avg_loss=avg_loss, avg_accuracy=avg_accuracy)
            logger.info("AUC for valication: %f", get_validation_result(model, X_val, y_val))

        from datetime import date
        today = date.today()
        torch.save(model.state_dict(), f"{today}_{output_model_file}")
# +
## +
## From baseline kernel
#
#    def calculate_overall_auc(df, model_name):
#        true_labels = df[TOXICITY_COLUMN] > 0.5
#        predicted_labels = df[model_name]
#
#        return metrics.roc_auc_score(true_labels, predicted_labels)
#
#    def power_mean(series, p):
#        total = sum(np.power(series, p))
#
#        return np.power(total / len(series), 1 / p)
#
#    def get_final_metric(bias_df, overall_auc, POWER=-5, OVERALL_MODEL_WEIGHT=0.25):
#        bias_score = np.average([
#            power_mean(bias_df[SUBGROUP_AUC], POWER),
#            power_mean(bias_df[BPSN_AUC], POWER),
#            power_mean(bias_df[BNSP_AUC], POWER)
#        ])
#
#        return (OVERALL_MODEL_WEIGHT * overall_auc) + ((1 - OVERALL_MODEL_WEIGHT) * bias_score)
#
#    SUBGROUP_AUC = 'subgroup_auc'
#    BPSN_AUC = 'bpsn_auc'  # stands for background positive, subgroup negative
#    BNSP_AUC = 'bnsp_auc'  # stands for background negative, subgroup positive
#
#    def compute_auc(y_true, y_pred):
#        try:
#            return metrics.roc_auc_score(y_true, y_pred)
#        except ValueError:
#            return np.nan
#
#    def compute_subgroup_auc(df, subgroup, label, model_name):
#        subgroup_examples = df[df[subgroup] > 0.5]
#
#        return compute_auc((subgroup_examples[label] > 0.5), subgroup_examples[model_name])
#
#    def compute_bpsn_auc(df, subgroup, label, model_name):
#        """Computes the AUC of the within-subgroup negative examples and the background positive examples."""
#        subgroup_negative_examples = df[(
#            df[subgroup] > 0.5) & (df[label] <= 0.5)]
#        non_subgroup_positive_examples = df[(
#            df[subgroup] <= 0.5) & (df[label] > 0.5)]
#        examples = subgroup_negative_examples.append(
#            non_subgroup_positive_examples)
#
#        return compute_auc(examples[label] > 0.5, examples[model_name])
#
#    def compute_bnsp_auc(df, subgroup, label, model_name):
#        """Computes the AUC of the within-subgroup positive examples and the background negative examples."""
#        subgroup_positive_examples = df[(
#            df[subgroup] > 0.5) & (df[label] > 0.5)]
#        non_subgroup_negative_examples = df[(
#            df[subgroup] <= 0.5) & (df[label] <= 0.5)]
#        examples = subgroup_positive_examples.append(
#            non_subgroup_negative_examples)
#
#        return compute_auc(examples[label] > 0.5, examples[model_name])
#
#    def compute_bias_metrics_for_model(dataset,
#                                       subgroups,
#                                       model,
#                                       label_col,
#                                       include_asegs=False):
#        """Computes per-subgroup metrics for all subgroups and one model."""
#        records = []
#
#        for subgroup in subgroups:
#            record = {
#                'subgroup': subgroup,
#                'subgroup_size': len(dataset[dataset[subgroup] > 0.5])
#            }
#            record[SUBGROUP_AUC] = compute_subgroup_auc(
#                dataset, subgroup, label_col, model)
#            record[BPSN_AUC] = compute_bpsn_auc(
#                dataset, subgroup, label_col, model)
#            record[BNSP_AUC] = compute_bnsp_auc(
#                dataset, subgroup, label_col, model)
#            records.append(record)
#
#        return pd.DataFrame(records).sort_values('subgroup_auc', ascending=True)
#
#
## +
#
#    test_df = train_df.tail(valid_size).copy()
#    train_df = train_df.head(num_to_load)
#    MODEL_NAME = 'model1'
#    test_df[MODEL_NAME] = torch.sigmoid(torch.tensor(valid_preds)).numpy()
#    TOXICITY_COLUMN = 'target'
#    bias_metrics_df
#    get_final_metric(
#        bias_metrics_df, calculate_overall_auc(test_df, MODEL_NAME))
## -



from fastai.basic_train import LearnerCallback
from fastai.basic_train import Learner

class SaveEveryNIterations(LearnerCallback):
    """Saves model after every N iterations
    We save all models with the same name as otherwise rather heavy models can quickly gobble up
    all available disk space.

    Usage:
    saver_callback = partial(SaveEveryNIterations, num_iterations=100,
                                              save_name="saved_every_100_iterations")
    learn = create_cnn(data, models.resnet18, callback_fns = [saver_callback])
    """
    def __init__(self, learn: Learner, num_iterations: int = 100, save_name=None, disable_callback:bool=False):
        """
        :param num_iterations: Saves model after every `num_iterations` iterations
        :param save_name: [optional] Filename to save model with
        """
        super().__init__(learn)
        self.num_iterations = num_iterations
        self.save_name = save_name
        if save_name is None:
            self.save_name = f'saved_every_{self.num_iterations}_iterations'

    def on_batch_end(self, iteration, **kwargs) -> None:
        if iteration % self.num_iterations == 0 and iteration != 0:
            #TODO : Report to Fastai : param names are different for saving model in language_model_learner
            if type(learn.model) == type(lambda x: x): # Checking if model is a function (changed by SkipNIterations0
                return
            self.learn.save(self.save_name)
            print(f"Model saved as {self.save_name} | Iteration : {iteration}")


class StopAfterNIterations(LearnerCallback):
    """Stops model after N iterations.
    Usage:
    stopper = partial(StopAfterNIterations, num_iterations = 17)
    learn = create_cnn(data, models.resnet18, callback_fns = [stopper])
    """
    def __init__(self, learn: Learner, num_iterations:int=100, disable_callback:bool=False):
        """
        :param num_iterations: Stops model after every `num_iterations` iterations
        """
        super().__init__(learn)
        self.num_iterations = num_iterations
        self.stop_training = False

    def on_batch_end(self, iteration, **kwargs) -> None:
        if iteration == self.num_iterations:
            print(f"Iteration {iteration} reached. Stopping Training")
            self.stop_training = True
            return {'stop_training': self.stop_training}

    def on_epoch_end(self, **kwargs) ->bool:
        if self.stop_training:
            print('Run learn.validate(learn.data.valid_dl) to see results')
            return {'stop_training': self.stop_training}


class GradientAccumulator(LearnerCallback):
    """Accumulates gradients over N iterations
    Usage:
    accumulator = partial(GradientAccumulator, num_iterations=100)
    learn = create_cnn(data, models.resnet18, callback_fns = [accumulator])
    """
    def __init__(self, learn: Learner, num_iterations: int = 4, disable_callback: bool = False):
        """
        :param num_iterations: Accumulate gradients over `num_iterations` iterations before taking an optimizer step
        """
        super().__init__(learn)
        self.num_iterations = num_iterations
        self.skipped_last_backprop = False

    def on_backward_end(self, iteration, **kwargs) -> None:
        if (iteration % self.num_iterations != 0) or (iteration == 0):
            self.skipped_last_backprop = True
            return {'skip_step': True, 'skip_zero': True}
        else:
            self.skipped_last_backprop = False

    def on_step_end(self, **kwargs):
        if self.skipped_last_backprop:
            return {'skip_zero': True}

    def on_epoch_end(self, **kwargs) ->bool:
        """Deals with the edge case of an epoch ending"""
        if self.skipped_last_backprop:
            self.learn.opt.step()
            self.learn.opt.zero_grad()


class ShowResutsEveryNIterations(LearnerCallback):
    """Shows model results after every N iterations

    Usage:
    results_callback = partial(SaveEveryNIterations, num_iterations=100)
    learn = create_cnn(data, models.resnet18, callback_fns = [results_callback])
    """
    def __init__(self, learn: Learner, num_iterations: int = 100, save_name=None, disable_callback:bool=False):
        """
        :param num_iterations: Show model resuts after every `num_iterations` iterations
        """
        super().__init__(learn)
        self.num_iterations = num_iterations

    def on_batch_end(self, iteration, **kwargs) -> None:
        if iteration % self.num_iterations == 0 and iteration != 0:
            self.learn.show_results()
            self.learn.model.train()


class SkipNIterations(LearnerCallback):
    """Skips first N iterations while training.
    Usage:
    skipper = partial(SkipNIterations, num_iterations = 20)
    learn = create_cnn(data, models.resnet18, callback_fns = [skipper])
    """
    def __init__(self, learn: Learner, num_iterations:int=100, disable_callback:bool=False):
        """
        :param num_batches: number of batches to skip in the beginning of training
        """
        super().__init__(learn)
        _order = 1
        self.num_iterations = num_iterations
        self.epoch_len = len(self.learn.data.train_dl)
        self.skip_epochs = self.num_iterations // self.epoch_len
        self.model = self.learn.model
        if self.skip_epochs > 0:
            self.num_iterations = self.num_iterations % self.epoch_len

    def faux_model(self, *args, **kwargs): # Replacement for Model
        if hasattr(self, 'output'):
            return self.output

    def on_batch_begin(self, epoch, **kwargs):
        if epoch <= self.skip_epochs - 1:
            return {'stop_epoch': True}

    def on_batch_end(self, num_batch:int, last_output, epoch, **kwargs):
        # We replace the model with a bogus function that returns the same output each time
        if num_batch == 1:
            self.output = last_output
            self.learn.model = self.faux_model
        if epoch <= self.skip_epochs - 1:
            return {'stop_epoch': True}
        if num_batch <= self.num_iterations - 1:
            return {'skip_validate': True}
        else:
            self.learn.model = self.model

    def on_backward_begin(self, num_batch:int, **kwargs):
        if num_batch <= self.num_iterations - 1:
            return {'skip_bwd': True}

    def on_backward_end(self, num_batch:int, **kwargs):
        if num_batch <= self.num_iterations - 1:
            return {'skip_step': True}

    def on_step_end(self, num_batch:int, **kwargs):
        if num_batch <= self.num_iterations - 1:
            return {'skip_step': True, 'skip_validate': True}
        else:
            return {'skip_validate':False} # Make sure we validate in the end



import os

import cv2
import numpy as np
from torch.utils.data import DataLoader, Dataset, sampler

from kaggle_runner.datasets.coders import run_length_decode
from kaggle_runner.datasets.transfomers import get_transforms


class SIIMDataset(Dataset):
    def __init__(self, df, fnames, data_folder, size, mean, std, phase):
        self.df = df
        self.root = data_folder
        self.size = size
        self.mean = mean
        self.std = std
        self.phase = phase
        self.transforms = get_transforms(phase, size, mean, std)
        self.gb = self.df.groupby("ImageId")
        self.fnames = fnames

    def __getitem__(self, idx):
        image_id = self.fnames[idx]
        df = self.gb.get_group(image_id)
        annotations = df[" EncodedPixels"].tolist()
        image_path = os.path.join(self.root, image_id + ".png")
        image = cv2.imread(image_path)
        mask = np.zeros([1024, 1024])
        if annotations[0] != "-1":
            for rle in annotations:
                mask += run_length_decode(rle)
        mask = (mask >= 1).astype("float32")  # for overlap cases
        augmented = self.transforms(image=image, mask=mask)
        image = augmented["image"]
        mask = augmented["mask"]
        return image, mask

    def __len__(self):
        return len(self.fnames)



import os
import subprocess

SEED = 1

FOCAL_LOSS_GAMMA = 0.0
FOCAL_LOSS_GAMMA_NEG_POS = 0.25
FOCAL_LOSS_BETA_NEG_POS = 1.0
# ALPHA = 0.91
ALPHA = 0.5

CONVERT_DATA = False
CONVERT_DATA_Y_NOT_BINARY = "convert_data_y_not_binary"
# given the pickle of numpy train data
CONVERT_TRAIN_DATA = "convert_train_data"
# given the pickle of numpy train data
CONVERT_ADDITIONAL_NONTOXIC_DATA = "CONVERT_ADDITIONAL_NONTOXIC_DATA"

EXCLUDE_IDENTITY_IN_TRAIN = True
TRAIN_DATA_EXCLUDE_IDENDITY_ONES = "TRAIN_DATA_EXCLUDE_IDENDITY_ONES"
DATA_ACTION_NO_NEED_LOAD_EMB_M = "DATA_ACTION_NO_NEED_LOAD_EMB_M"

NEG_RATIO = 1 - 0.05897253769515213
Data_Folder = '/home/'
LOAD_BERT_DATA = False

RIPDB = os.environ.get("RIPDB") == 'true'
DEBUG = os.environ.get("DEBUG") == 'true'
INTERACTIVE = False
try:
	__note = subprocess.check_output('pgrep -f "jupyter notebook"', shell=True)

	if __note.strip() != "":
		INTERACTIVE = True
except:
	pass



from enum import Enum


class KernelRunningState(Enum):
    INIT_DONE = 1
    PREPARE_DATA_DONE = 2
    TRAINING_DONE = 3
    EVL_DEV_DONE = 4
    SAVE_SUBMISSION_DONE = 5

    @staticmethod
    def string(state_int_value):
        names = [
            "INIT_DONE",
            "PREPARE_DATA_DONE",
            "TRAINING_DONE",
            "EVL_DEV_DONE",
            "SAVE_SUBMISSION_DONE",
        ]
        if state_int_value is not None:
            return names[state_int_value]
        else:
            return ""



from glob import glob
import time
import os

import torch
import torch.backends.cudnn as cudnn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from fastai import *
from fastai.core import *
from fastai.torch_core import *
from fastai.basic_data import *
from fastai.basic_train import LearnerCallback, Learner
from catalyst.data.sampler import DistributedSamplerWrapper, BalanceClassSampler
from datetime import datetime

from kaggle_runner import logger
from kaggle_runner.kernels.fastai_kernel import FastAIKernel
from kaggle_runner.metrics.meters import AverageMeter, RocAucMeter
from kaggle_runner.data_providers import provider
from kaggle_runner.logs import metric_get_log
from kaggle_runner.losses import MixedLoss
from kaggle_runner.metrics.meters import Meter
from kaggle_runner.optimizers import RAdam


def _change_dl(k, dl, shuffle):
    old_dl = dl
    train_sampler = DistributedSamplerWrapper(
        sampler=BalanceClassSampler(
            labels=k.train_dataset.get_labels(), mode="downsampling"),
        num_replicas=8,  # xm.xrt_world_size(),
        rank=0,  # xm.get_ordinal(), it only get 1/8 data ....
        shuffle=True
    )
    train_loader = torch.utils.data.DataLoader(
        k.train_dataset,
        batch_size=k.config.batch_size,
        sampler=train_sampler,
        pin_memory=True,
        drop_last=True,
        num_workers=k.config.num_workers,
    )
    new_dl = train_loader

    return old_dl, new_dl, train_sampler


def _change_dl_val(k, dl, shuffle):
    old_dl = dl
    validation_sampler = torch.utils.data.distributed.DistributedSampler(
        k.validation_dataset,
        num_replicas=8,  # xm.xrt_world_size(),
        rank=0,  # xm.get_ordinal(),
        shuffle=False
    )
    validation_loader = torch.utils.data.DataLoader(
        k.validation_dataset,
        batch_size=k.config.batch_size,
        sampler=validation_sampler,
        pin_memory=False,
        drop_last=False,
        num_workers=k.config.num_workers
    )

    return old_dl, validation_loader, validation_sampler


class Trainer:
    """This class takes care of training and validation of our model"""

    def __init__(self, model, data_folder, df_path):
        self.fold = 1
        self.total_folds = 5
        self.num_workers = 4
        self.batch_size = {"train": 4, "val": 4}
        self.accumulation_steps = 32 // self.batch_size["train"]
        self.lr = 5e-4
        self.num_epochs = 32
        self.best_loss = float("inf")
        self.phases = ["train", "val"]
        self.device = torch.device("cpu")
        # self.device = torch.device("cuda:0")
        # torch.set_default_tensor_type("torch.cuda.FloatTensor")
        self.net = model
        self.criterion = MixedLoss(10.0, 2.0)
        # self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)
        self.optimizer = RAdam(model.parameters(), lr=self.lr)

        self.scheduler = ReduceLROnPlateau(
            self.optimizer, mode="min", patience=3, verbose=True
        )
        self.net = self.net.to(self.device)
        cudnn.benchmark = True
        self.dataloaders = {
            phase: provider(
                fold=1,
                total_folds=5,
                data_folder=data_folder,
                df_path=df_path,
                phase=phase,
                size=512,
                mean=(0.485, 0.456, 0.406),
                std=(0.229, 0.224, 0.225),
                batch_size=self.batch_size[phase],
                num_workers=self.num_workers,
            )

            for phase in self.phases
        }
        self.losses = {phase: [] for phase in self.phases}
        self.iou_scores = {phase: [] for phase in self.phases}
        self.dice_scores = {phase: [] for phase in self.phases}

    def forward(self, images, targets):
        images = images.to(self.device)
        masks = targets.to(self.device)
        outputs = self.net(images)
        loss = self.criterion(outputs, masks)

        return loss, outputs

    def iterate(self, epoch, phase):
        meter = Meter(phase, epoch)
        start = time.strftime("%H:%M:%S")
        print(f"Starting epoch: {epoch} | phase: {phase} | ⏰: {start}")
        batch_size = self.batch_size[phase]
        self.net.train(phase == "train")
        dataloader = self.dataloaders[phase]
        running_loss = 0.0
        total_batches = len(dataloader)
        #         tk0 = tqdm(dataloader, total=total_batches)
        self.optimizer.zero_grad()

        for itr, batch in enumerate(dataloader):
            images, targets = batch
            loss, outputs = self.forward(images, targets)
            loss = loss / self.accumulation_steps

            if phase == "train":
                loss.backward()

                if (itr + 1) % self.accumulation_steps == 0:
                    self.optimizer.step()
                    self.optimizer.zero_grad()
            running_loss += loss.item()
            outputs = outputs.detach().cpu()
            meter.update(targets, outputs)
        #             tk0.set_postfix(loss=(running_loss / ((itr + 1))))
        epoch_loss = (running_loss * self.accumulation_steps) / total_batches
        dice, iou = metric_get_log(phase, epoch, epoch_loss, meter, start)
        self.losses[phase].append(epoch_loss)
        self.dice_scores[phase].append(dice)
        self.iou_scores[phase].append(iou)
        torch.cuda.empty_cache()

        return epoch_loss

    def start(self):
        for epoch in range(self.num_epochs):
            self.iterate(epoch, "train")
            state = {
                "epoch": epoch,
                "best_loss": self.best_loss,
                "state_dict": self.net.state_dict(),
                "optimizer": self.optimizer.state_dict(),
            }
            val_loss = self.iterate(epoch, "val")
            self.scheduler.step(val_loss)

            if val_loss < self.best_loss:
                print("******** New optimal found, saving state ********")
                state["best_loss"] = self.best_loss = val_loss
                torch.save(state, "./model.pth")
            print()


class GPUTrainer(LearnerCallback):
    def __init__(self, learn: Learner, k: FastAIKernel):
        super().__init__(learn)
        self.k = k

    def on_train_begin(self, **kwargs: Any) -> None:
        #self.device = xm.xla_device(devkind='CPU')
        self.device = torch.device("cuda")
        self.learn.model = self.learn.model.to(self.device)
        #self.learn.data.add_tfm(partial(batch_to_device,device=self.device))
        self.old_sampler_train_dl, self.data.train_dl, self.train_sampler = _change_dl(self.k,
                                                                                       self.data.train_dl, shuffle=True)
        self.old_sampler_valid_dl, self.data.valid_dl, self.valid_sampler = _change_dl_val(self.k,
                                                                                           self.data.valid_dl, shuffle=False)

        #self.learn.data.add_tfm(partial(batch_to_device,device=self.device))
        self.learn.data.train_dl = DeviceDataLoader(
            self.data.train_dl, device=self.device)
        self.learn.data.valid_dl = DeviceDataLoader(
            self.data.valid_dl, device=self.device)
        #self.learn.data.train_dl = pl.ParallelLoader(self.data.train_dl, [self.device]).per_device_loader(self.device)
        #self.learn.data.valid_dl = pl.ParallelLoader(self.data.valid_dl, [self.device]).per_device_loader(self.device)
        #self.learn.data.train_dl.dataset = None #self.old_train_dl.dataset
        #self.learn.data.valid_dl.dataset = None #self.old_train_dl.dataset

    def on_backward_end(self, **kwargs: Any) -> None:
        #xm.optimizer_step(self.learn.opt.opt, barrier=True)
        pass



import gc
import importlib
import os
import random
import unittest

import numpy as np
import pytest
import torch
from torch.autograd import Variable

# import modelTester
from kaggle_runner.runners import runner
from kaggle_runner.data_providers import provider
from kaggle_runner.kernels import PSKernel, kernel, pytorchKernel
from kaggle_runner.kernels.KernelRunningState import KernelRunningState
# from kernel import Kernel
from kaggle_runner.utils import kernel_utils

log_args = {
    "size": 384,
    "network": "intercept",
    "AMQPURL": "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
    "seed": 19999,
}
r = runner.Runner(
    log_args["network"],
    log_args["AMQPURL"],
    size=log_args["size"],
    seed=log_args["seed"],
)
r._attach_data_collector("")
LOGGER = r.logger


@pytest.fixture(scope="session")
def args():
    return {
        "size": 384,
        "network": "intercept",
        "AMQPURL": "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
        "seed": 19999,
    }


@pytest.fixture(scope="session")
def mq_logger():
    log_args = {
        "size": 384,
        "network": "intercept",
        "AMQPURL": "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
        "seed": 19999,
    }
    r = runner.Runner(
        log_args["network"],
        log_args["AMQPURL"],
        size=log_args["size"],
        seed=log_args["seed"],
    )
    r._attach_data_collector("")
    _mq_logger = r.logger
    return _mq_logger


class TestModelBasic:
    ModelList = []

    def test_creation(self):
        pass

    def test_load_model(self):
        pass

    def test_eval(self):
        pass

    def test_dataloaders(self):
        """test_dataloaders uses codes from unet-with-se-resnet ..., it uses
        provider to generate dataloader"""
        train_rle_path = "../input/mysiim/train-rle.csv"
        data_folder = "../input/siimpng/siimpng/train_png"
        dataloader = provider(
            fold=0,
            total_folds=5,
            data_folder=data_folder,
            df_path=train_rle_path,
            phase="train",
            size=512,
            mean=(0.485, 0.456, 0.406),
            std=(0.229, 0.224, 0.225),
            batch_size=16,
            num_workers=2,
        )

        batch = next(iter(dataloader))  # get a batch from the dataloader
        images, masks = batch

        # plot some random images in the `batch`
        idx = random.choice(range(16))
        # plt.imshow(images[idx][0], cmap="bone")
        # plt.imshow(masks[idx][0], alpha=0.2, cmap="Reds")
        # plt.show()
        if len(np.unique(masks[idx][0])) == 1:  # only zeros
            print("Chosen image has no ground truth mask, rerun the cell")


class TestMQLogger:
    def test_runner_create(self, args):
        r = runner.Runner(
            args["network"], args["AMQPURL"], size=args["size"], seed=args["seed"]
        )
        assert r.AMQPURL is not None

    def test_runner_logger_use_elsewhere(self, args):
        r = runner.Runner(
            args["network"], args["AMQPURL"], size=args["size"], seed=args["seed"]
        )
        assert r.AMQPURL is not None
        r._attach_data_collector("")
        assert r.logger is not None
        r.logger.debug("use elsewhere")


@pytest.yield_fixture(scope="session")
def session_thing(mq_logger):
    mq_logger.debug("constructing session thing")
    yield
    mq_logger.debug("destroying session thing")


@pytest.yield_fixture
def testcase_thing(mq_logger):
    mq_logger.debug("constructing testcase thing")
    yield
    mq_logger.debug("destroying testcase thing")


class TestPSKernel:
    # this function will run before every test. We re-initialize group in this
    # function. So for every test, new group is used.
    @classmethod
    def setup_class(cls):
        gc.enable()
        importlib.reload(kernel_utils)
        importlib.reload(kernel)
        importlib.reload(PSKernel)
        importlib.reload(pytorchKernel)
        # importlib.reload(modelTester)
        log_args = {
            "size": 384,
            "network": "intercept",
            "AMQPURL": "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
            "seed": 19999,
        }
        r = runner.Runner(
            log_args["network"],
            log_args["AMQPURL"],
            size=log_args["size"],
            seed=log_args["seed"],
        )
        r._attach_data_collector("")
        cls.logger = r.logger
        cls.logger.debug("Good day~")

    @classmethod
    def teardown_class(cls):
        cls.logger.debug("Keep happy~")

    def setup_method(self, method):
        self.logger.debug("setup for method %s", method)

    def teardown_method(self, method):
        self.logger.debug("teardown method %s", method)
        gc.collect()

    def test_class(self, mq_logger):
        ps_kernel = PSKernel.PS(mq_logger)
        # set_trace()
        assert len(ps_kernel.model_metrics) == 0

    @pytest.mark.skip("take too long to test, just skip")
    def test_prepare_data(self, mq_logger):
        ps_kernel = PSKernel.PS(mq_logger)
        ps_kernel.run(
            end_stage=KernelRunningState.PREPARE_DATA_DONE, dump_flag=False
        )  # will also analyze data
        assert ps_kernel.train_X is not None
        assert len(ps_kernel.train_X) == len(ps_kernel.train_Y)
        # self.assertIsNotNone(ps_kernel.test_X)  // don't care this now
        assert ps_kernel.dev_X is not None
        assert len(ps_kernel.dev_X) == len(ps_kernel.dev_Y)

    @pytest.mark.skip("take too long to test, just skip")
    def test_dump_load_continue(self, mq_logger):
        ps_kernel = PSKernel.PS(mq_logger)
        ps_kernel.run(end_stage=KernelRunningState.TRAINING_DONE)
        assert ps_kernel._stage == KernelRunningState.TRAINING_DONE

        kernel_load_back = kernel.KaggleKernel._load_state(logger=mq_logger)
        assert kernel_load_back._stage == KernelRunningState.TRAINING_DONE
        kernel_load_back.run()
        assert kernel_load_back._stage == KernelRunningState.SAVE_SUBMISSION_DONE

    @pytest.mark.skip("take too long to test, just skip")
    def test_train(self, mq_logger):
        # kernel_load_back = kernel.KaggleKernel._load_state(
        #     KernelRunningState.PREPARE_DATA_DONE, logger=mq_logger
        # )
        ps_kernel = PSKernel.PS(mq_logger)
        ps_kernel.run(end_stage=KernelRunningState.TRAINING_DONE)
        assert ps_kernel.model is not None

    @pytest.mark.skip("take too long to test, just skip")
    def test_read_tf(self, mq_logger):
        k = PSKernel.PS(mq_logger)
        k._recover_from_tf()
        # k.run(start_stage=KernelRunningState.PREPARE_DATA_DONE,
        # end_stage=KernelRunningState.TRAINING_DONE)
        assert k.ds is not None

    @pytest.mark.skip("take too long to test, just skip")
    def test_convert_tf_from_start(self, mq_logger):  # won't work
        ps_kernel = PSKernel.PS(mq_logger)
        ps_kernel.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)
        assert os.path.isfile("train_dev.10.tfrec")

    # def test_convert_tf(self, mq_logger):
    #    kernel_withdata
    #    = kernel.KaggleKernel._load_state(KernelRunningState.PREPARE_DATA_DONE)
    #    k = PSKernel.PS(mq_logger)
    #    k._clone_data(kernel_withdata)
    #    k.after_prepare_data_hook()

    #    self.assertTrue(os.path.isfile('train.tfrec'))


@pytest.mark.skip("take too long to test, just skip")
class TestPytorchKernel:
    @classmethod
    def setup_class(cls):
        gc.enable()
        importlib.reload(kernel_utils)
        importlib.reload(kernel)
        importlib.reload(PSKernel)
        importlib.reload(pytorchKernel)
        # importlib.reload(modelTester)
        log_args = {
            "size": 384,
            "network": "intercept",
            "AMQPURL": "amqp://drdsfaew:QrBHPPxbsd8IuIxKrCnX3-RGoLKaFhYI@termite.rmq.cloudamqp.com/drdsfaew",
            "seed": 19999,
        }
        r = runner.Runner(
            log_args["network"],
            log_args["AMQPURL"],
            size=log_args["size"],
            seed=log_args["seed"],
        )
        r._attach_data_collector("")
        cls.logger = r.logger

        cls.logger.debug("Good day~")

    @classmethod
    def teardown_class(cls):
        cls.logger.debug("Keep happy~")

    def setup_method(self, method):
        self.logger.debug("setup for method %s", method)

    def teardown_method(self, method):
        self.logger.debug("teardown method %s", method)
        gc.collect()

    def test_pytorch_data_aug(self, mq_logger):
        self._prepare_data(mq_logger)

        k = pytorchKernel.PS_torch(mq_logger)
        k._debug_less_data = True
        k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)
        # k.load_state_data_only(KernelRunningState.PREPARE_DATA_DONE)
        # k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE,
        # dump_flag=True)  # will also analyze data
        k.data_loader.dataset._test_(1)
        k.data_loader.dataset._test_(2)  # test should choose small idx, as
        # batch might be small
        # k.run()  # dump not working for torch
        assert k is not None

    def _prepare_data(self, mq_logger):
        _stage = KernelRunningState.PREPARE_DATA_DONE
        data_stage_file_name = "run_state_%s.pkl" % _stage
        if not os.path.isfile(data_stage_file_name):
            self._pytorch_starter_dump(mq_logger)

    @pytest.mark.skip("data dump not avaliable")
    def test_pytorch_FL_in_model_early_stop(self, mq_logger):
        self._prepare_data(mq_logger)
        kernel_load_back = pytorchKernel.PS_torch(mq_logger)

        kernel_load_back.load_state_data_only(
            KernelRunningState.PREPARE_DATA_DONE)
        kernel_load_back.build_and_set_model()
        kernel_load_back.train_model()

    @pytest.mark.skip("we train more")
    def test_pytorch_cv_train_dev(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)

        k._debug_less_data = True
        k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)
        k.logger.debug("data done")

        k.build_and_set_model()
        k.num_epochs = 1
        k.logger.debug("start train one epoch")
        k.train_model()
        k.logger.debug("end train one epoch")
        assert True

    def test_pytorch_cv_train_more(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)

        # k._debug_less_data = True
        k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)

        k.build_and_set_model()
        k.logger.debug("start train 5 epochs")
        k.num_epochs = 5
        k.train_model()

    def test_pytorch_cv_data_prepare(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)
        k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)

        # k2 = pytorchKernel.PS_torch(mq_logger)
        # k2.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)

        l = len(k.data_loader)
        l_dev = len(k.data_loader_dev)
        ratio = l / l_dev

        assert ratio > 3.5  # around 0.8/0.2
        assert ratio < 4.5

    def test_pytorch_focal_loss_func(self, mq_logger):
        inputs = torch.tensor(
            [
                [0.0, 0.2, 0.4, 0.0, 0.2, 0.4, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
                [0.0, 0.2, 0.4, 0.0, 0.2, 0.4, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
            ]
        )
        targets = torch.tensor(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
            ]
        )

        FL = pytorchKernel.FocalLoss(gamma=2)
        FL_normal_CE = pytorchKernel.FocalLoss(gamma=0)
        alpha_for_pos = 0.75
        FL_alpha_balance = pytorchKernel.FocalLoss(
            gamma=0, alpha=alpha_for_pos
        )  # alpha for positive class weights

        print("----inputs----")
        print(inputs)
        print("---target-----")
        print(targets)
        losses = []
        grads = []

        for loss_func in [FL, FL_normal_CE, FL_alpha_balance, FL_normal_CE]:
            inputs_fl = Variable(inputs.clone(), requires_grad=True)
            targets_fl = Variable(targets.clone())
            fl_loss = loss_func(inputs_fl, targets_fl)
            print(fl_loss.data)
            fl_loss.backward()
            print(inputs_fl.grad.data)
            losses.append(fl_loss.data.numpy())
            grads.append(inputs_fl.grad.data.numpy())
        assert losses[-1] == losses[1]
        if alpha_for_pos >= 0.5:
            assert losses[2] < losses[1]
        else:
            assert losses[2] > losses[1]

        assert (grads[2][0, 1] / grads[1][0, 1]) == (1 - alpha_for_pos) / 0.5

    @pytest.mark.skip("won't work for pytorch")
    def _pytorch_starter_dump(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)
        k.run(
            end_stage=KernelRunningState.PREPARE_DATA_DONE, dump_flag=True
        )  # will also analyze data
        # kernel_load_back = pytorchKernel.PS_torch(mq_logger)

        # kernel_load_back.load_state_data_only(KernelRunningState.PREPARE_DATA_DONE)
        # kernel_load_back.run(end_stage=KernelRunningState.TRAINING_DONE)

    @pytest.mark.skip("won't work for pytorch")
    def test_pytorch_starter_load(self, mq_logger):
        kernel_load_back = pytorchKernel.PS_torch(mq_logger)

        kernel_load_back.load_state_data_only(KernelRunningState.TRAINING_DONE)
        kernel_load_back.load_model_weight()

    @pytest.mark.skip("won't work for pytorch")
    def test_pytorch_starter_load_then_train(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)

        # will also analyze data
        k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE)
        # k._debug_continue_training = True
        k.load_model_weight_continue_train()
        # kernel_load_back.run(end_stage=KernelRunningState.TRAINING_DONE)

    @pytest.mark.skip("won't work for pytorch")
    def test_pytorch_starter_load_then_submit(self, mq_logger):
        kernel_load_back = pytorchKernel.PS_torch(mq_logger)

        kernel_load_back.load_state_data_only(KernelRunningState.TRAINING_DONE)
        kernel_load_back.load_model_weight()
        kernel_load_back.run()
        # kernel_load_back.run(end_stage=KernelRunningState.TRAINING_DONE)

    def test_pytorch_dataset_mean_std(self, mq_logger):
        k = pytorchKernel.PS_torch(mq_logger)
        # k.run(end_stage=KernelRunningState.PREPARE_DATA_DONE,
        # dump_flag=True)  # will also analyze data
        k._debug_less_data = True
        k.run(
            end_stage=KernelRunningState.PREPARE_DATA_DONE, dump_flag=False
        )  # dump not working for torch
        k.pre_train()
        assert k.img_mean is not None

    # def test_tf_model_zoo(self, mq_logger):
    #     t = modelTester.TF_model_zoo_tester()
    #     t.run_logic()

    # def test_tf_model_zoo_model(self, mq_logger):
    #     t = modelTester.TF_model_zoo_tester()
    #     t.load_model()
    #     assert t.model is not None

    # def test_tf_model_zoo_model(self, mq_logger):
    #     t = modelTester.TF_model_zoo_tester()
    #     t.set_model("mask_rcnn_resnet101_atrous_coco_2018_01_28")
    #     t.run_prepare()
    #     # t.check_graph()
    #     t.run_test()  # result is great!!!
    #     assert t.detection_graph is not None

    @pytest.mark.skip()
    def test_analyze_RPN(self, mq_logger):
        assert False

    @pytest.mark.skip()
    def test_analyze_predict_error(self, mq_logger):
        assert False

    @pytest.mark.skip()
    def test_analyze_predict_score_threshold(self, mq_logger):
        ts = np.exp([0.5, 0.6, 0.7])
        # for t in ts:
        #     check_predict_statistics()
        assert False

    @pytest.mark.skip()
    def test_TTA(self, mq_logger):
        assert False

    @pytest.mark.skip()
    def test_L_loss(self, mq_logger):
        assert False


if "__main__" == __name__:
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kaggle-runner",
    version="0.1.7",
    description="Run kaggle kernels, for fast model prototyping.",
    url="http://github.com/pennz/kaggle_runner",
    author="pennz",
    author_email="pengyuzhou.work@gmail.com",
    license="MIT",
    zip_safe=False,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "slug",
        "parse",
        "python_logging_rabbitmq",
        "kaggle",
        "colorama"
    ],
)
